from typing import *

from base_aux.testplans.devices_base import *
from base_aux.testplans.tc_types import *
from base_aux.testplans.tc import *

from base_aux.valid.m1_valid_base import *
from base_aux.valid.m2_valid_derivatives import *
from base_aux.valid.m3_valid_chains import *
from base_aux.base_values.m5_value_valid1_eq import *


# =====================================================================================================================
# class Enum_TcGroup(NestEq_EnumAdj):
#     NONE = None


# =====================================================================================================================
class _Base_TcAtc(Base_TestCase):
    # SETTINGS -------------------------------
    ATC_VOUT: int | None = None

    # INTERNAL AUX ---------------------------
    @classmethod
    @property
    def _EQ_CLS__VALUE(cls) -> str:
        """
        GOAL
        ----
        REDEFINE TO USE AS CMP VALUE
        """
        return f"Atc{cls.ATC_VOUT}"

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def startup__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        result_chain = ValidChains(
            chains=[
                ValidBreak(not cls.ATC_VOUT),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].connect__only_if_address_resolved,
                    name="ATC.connect__only_if_address_resolved",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].GET,
                    args__value="VSNS1",
                    validate_link=1,
                    name="GET",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].GET,
                    args__value="VSNS2",
                    validate_link=1,
                    name="GET",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].SET,
                    kwargs__value={"VINPIN": 230, "__timeout": 10},
                    validate_link="OK",
                    name="SET",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].SET,
                    kwargs__value={"VOUTPIN": cls.ATC_VOUT, "__timeout": 10},
                    validate_link="OK",
                    name="SET",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].SET,
                    kwargs__value={"PWR": "ON"},
                    validate_link="OK",
                    name="SET",
                ),
                ValidSleep(0.5),
                # Valid(
                #     value_link=cls.STAND.DEV_LINES.ATC[0].GET,
                #     args__value="VOUT",
                #     validate_link=lambda source: ValidAux.check_lege(source, cls.ATC_VOUT * 0.9, cls.ATC_VOUT * 1.1),
                #     name="GET+check_lege",
                # ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0]._buffers_clear__read,
                    validate_link=lambda value: value is None,
                    name="_buffers_clear__read",
                ),
            ],
        )
        return result_chain

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def teardown__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        result_chain = ValidChains(
            chains=[
                ValidBreak(not cls.ATC_VOUT),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].connect__only_if_address_resolved,
                    name="ATC.connect__only_if_address_resolved",
                ),
                Valid(
                    value_link=cls.STAND.DEV_LINES.ATC[0].reset,
                    validate_link=lambda value: value is None,
                    name="RESET",
                ),
            ]
        )
        return result_chain


# =====================================================================================================================
class Base_TcAtcPtb(_Base_TcAtc):
    ATC_VOUT: int | None = 0
    PTB_SET_EXTON: bool = False
    PTB_SET_HVON: bool = False
    PTB_SET_PSON: bool = False

    ASYNC = True
    INFO_STR__ADD_ATTRS = ["ATC_VOUT", "PTB_SET_EXTON", "PTB_SET_HVON", "PTB_SET_PSON"]

    _DESCRIPTION = "[base] for testing params in different states"

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    @property
    def DESCRIPTION_prefix(cls) -> str:
        result = "["
        result += f"Atc{cls.ATC_VOUT},"
        result += f"Ext{'On' if cls.PTB_SET_EXTON else 'Off'},"
        result += f"Hv{'On' if cls.PTB_SET_HVON else 'Off'},"
        result += f"Ps{'On' if cls.PTB_SET_PSON else 'Off'}"
        result += "]"
        # result += "\n"    # dont use here! use in _DESCR* directly!
        return result

    @classmethod
    @property
    def DESCRIPTION(cls) -> str:
        return cls.DESCRIPTION_prefix + cls._DESCRIPTION

    # -----------------------------------------------------------------------------------------------------------------
    def expect__psu_may_start__hv(self) -> bool:
        """
        show if PSU must run correctly
        """
        return all([150 < self.ATC_VOUT < 260, self.PTB_SET_HVON])

    def expect__psu_may_start__hv_or_ext(self) -> bool:
        """
        show if PSU must run correctly
        """
        return self.expect__psu_may_start__hv() or self.PTB_SET_EXTON

    def expect__psu_started_hv__ok(self) -> bool:
        """
        show if PSU must run correctly
        """
        return all([150 < self.ATC_VOUT < 260, self.PTB_SET_HVON, self.PTB_SET_PSON])

    def expect__psu_started_hv__fail(self) -> bool:
        """
        WHY NOT - USE JUST SIMPLE NOT*
        ------------------------------
        cause i need link!
        """
        return not self.expect__psu_started_hv__ok()

    # -----------------------------------------------------------------------------------------------------------------
    def steps__check_params(self) -> list[Valid]:
        """
        GOAL
        ----
        separate list Valid steps to use in any TC
        :return:
        """
        result = [
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="VINOK",
                    validate_link=1 if self.expect__psu_may_start__hv() else 0,
                    name="GET",
                ),
                ValidFailContinue(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="PWOK",
                    validate_link=1 if self.expect__psu_started_hv__ok() else 0,
                    name="GET",
                ),

                # HV ----
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="VIN",
                    validate_link=EqValid_LGTE_NumParsedSingle(
                        **(dict(ge=210, le=230)
                        if self.expect__psu_may_start__hv()
                        else
                        dict(ge=0, le=0.1))
                    ).resolve,
                    name="GET+valid diapason",
                ),

                # LV ----
                ValidFailContinue(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="V12S",
                    validate_link=EqValid_LGTE_NumParsedSingle(
                        **(dict(ge=11, le=13)
                        if self.expect__psu_may_start__hv_or_ext()
                        else
                        dict(ge=0, le=0.1))
                    ).resolve,
                    name="GET+valid diapason",
                ),
                ValidFailContinue(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="V12M",
                    validate_link=EqValid_LGTE_NumParsedSingle(
                        **(dict(ge=11, le=13)
                        if self.expect__psu_may_start__hv_or_ext() and self.PTB_SET_PSON
                        else
                        dict(ge=0, le=0.1))
                    ).resolve,
                    name="GET+valid diapason",
                ),

                # currents ----
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="IIN",
                    validate_link=EqValid_LGTE_NumParsedSingle(ge=0, le=0.1).resolve,
                    name="GET+valid diapason",
                ),
                ValidFailContinue(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="IOUT",
                    validate_link=EqValid_LGTE_NumParsedSingle(ge=0, le=0.1).resolve,
                    name="GET+valid diapason",
                ),
            ]
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def run__wrapped(self) -> TYPING__RESULT_W_EXC:
        # BY DEFAULT PARAMS WOULD CHECK!!!
        result_chain = ValidChains(
            chains=self.steps__check_params()
        )
        return result_chain

    # =================================================================================================================
    def startup__wrapped(self) -> TYPING__RESULT_W_EXC:
        result_chain = ValidChains(
            [
                # ValidBreak(not any([self.PTB_SET_EXTON, self.PTB_SET_HVON, self.PTB_SET_PSON])),

                Valid(
                    value_link=self.DEV_COLUMN.DUT.connect__only_if_address_resolved,
                    name="DUT.connect__only_if_address_resolved",
                ),

                # LED ----------------------------------------------------
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.SET,
                    kwargs__value={"LED2": "ON"},
                    validate_link="OK",
                    name="SET",
                ),

                # PTB_SET_EXTON ----------------------------------------------------
                Valid(
                    skip_link=not self.PTB_SET_EXTON,
                    value_link=self.DEV_COLUMN.DUT.SET,
                    kwargs__value={"EXT12": "ON"},
                    validate_link="OK",
                    name="SET",
                ),
                ValidSleep(0.5, skip_link=not self.PTB_SET_EXTON),

                # HV0 --------------------------------------------------------------
                Valid(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="HV0",
                    validate_link=1 if self.ATC_VOUT else 0,
                    name="GET",
                ),

                # PTB_SET_HVON
                Valid(
                    skip_link=not self.PTB_SET_HVON or self.PTB_SET_EXTON,
                    value_link=self.DEV_COLUMN.DUT.SET,
                    kwargs__value={"HV": "ON"},
                    validate_link="OK",
                    name="SET",
                ),
                ValidSleep(0.5, skip_link=not self.PTB_SET_HVON),

                # HV1 -----------
                Valid(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="HV1",
                    validate_link=1 if self.expect__psu_started_hv__ok() else 0,
                    name="GET",
                ),

                # PTB_SET_PSON ------------------------------------------------------
                Valid(
                    skip_link=not self.PTB_SET_PSON,
                    value_link=self.DEV_COLUMN.DUT.SET,
                    kwargs__value={"PSON": "ON", "__timeout": 10},
                    validate_link="OK" if self.expect__psu_may_start__hv_or_ext() else "ERR",
                    name="ON",
                ),
                ValidSleep(0.5, skip_link=not self.PTB_SET_PSON),
            ],
        )
        return result_chain

    # -----------------------------------------------------------------------------------------------------------------
    def teardown__wrapped(self) -> TYPING__RESULT_W_EXC:
        result_chain = ValidChains(
            chains=[
                # ValidBreak(not any([self.PTB_SET_EXTON, self.PTB_SET_HVON, self.PTB_SET_PSON])),

                Valid(
                    value_link=self.DEV_COLUMN.DUT.connect__only_if_address_resolved,
                    name="DUT.connect__only_if_address_resolved",
                ),
                Valid(
                    value_link=self.DEV_COLUMN.DUT.reset,
                    validate_link=lambda value: value is None,
                    name="RESET",
                ),

            ]
        )
        return result_chain


# =====================================================================================================================



from base_aux.base_nest_dunders.m3_calls import *
from base_aux.aux_text.m1_text_aux import *
from base_aux.base_nest_dunders.m1_init1_source import *

from base_aux.base_types.m0_static_types import TYPES


# =====================================================================================================================
class _NumParsedSingle(NestInit_Source, NestCall_Resolve):
    SOURCE: TYPES.NUMBER = None

    _numtype: EnumAdj_NumType = EnumAdj_NumType.BOTH

    def _init_post(self) -> None | NoReturn:
        self.SOURCE = TextAux(self.SOURCE).parse__number_single(num_type=self._numtype)

    def resolve(self) -> TYPES.NUMBER | None:
        return self.SOURCE


# ---------------------------------------------------------------------------------------------------------------------
@final
class NumParsedSingle(_NumParsedSingle):
    pass


@final
class NumParsedSingleInt(_NumParsedSingle):
    _numtype: EnumAdj_NumType = EnumAdj_NumType.INT


@final
class NumParsedSingleFloat(_NumParsedSingle):
    _numtype: EnumAdj_NumType = EnumAdj_NumType.FLOAT


# =====================================================================================================================



"""
GOAL
----
collect all obvious variants of code base_types

USEFUL IDEAS
------------
- in tests
- could be clearly used in docstrings without needless defining
    assert get_bool(LAMBDA_EXC) is False
"""
# =====================================================================================================================
from typing import *
import asyncio


# =====================================================================================================================
RANGE0: Iterable[int] = range(0)
RANGE1: Iterable[int] = range(1)
RANGE2: Iterable[int] = range(2)
RANGE3: Iterable[int] = range(3)


# =====================================================================================================================
class VALUES_BLANK__ELEMENTARY_SINGLE:
    NONE: None = None
    BOOL: bool = False
    INT: int = 0
    FLOAT: float = 0.0
    STR: str = ""         # that is why i create all others BLANKS_*
    BYTES: bytes = b""


class VALUES_BLANK__ELEMENTARY_COLLECTION:
    LIST: list = []
    TUPLE: tuple = ()
    DICT: dict = {}
    SET: set = set()
    RANGE: Iterable = RANGE0


@final
class VALUES_BLANK(VALUES_BLANK__ELEMENTARY_SINGLE, VALUES_BLANK__ELEMENTARY_COLLECTION):
    """
    GOAL
    ----
    keep all variants in one object with ability to iterate in bunch checks!
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
class VALUES_NOT_BLANK__ELEMENTARY_SINGLE:
    # NONE: None = None
    BOOL: bool = True
    INT: int = 1
    FLOAT: float = 1.2
    STR: str = "str"         # that is why i create all others BLANKS_*
    BYTES: bytes = b"bytes"


class VALUES__ELEMENTARY_SINGLE(VALUES_NOT_BLANK__ELEMENTARY_SINGLE):
    NONE: None = None


class VALUES_NOT_BLANK__ELEMENTARY_COLLECTION:
    LIST: list = [1, 2, ]
    TUPLE: tuple = (1, 2, )
    DICT: dict = {1: 11}
    SET: set = (1, 2, )
    RANGE: Iterable = RANGE1


@final
class VALUES_NOT_BLANK(VALUES_NOT_BLANK__ELEMENTARY_SINGLE, VALUES_NOT_BLANK__ELEMENTARY_COLLECTION):
    pass


# =====================================================================================================================
COMPR_TUPLE_GEN: Generator = (i for i in range(3))
COMPR_LIST: list = [i for i in range(3)]
COMPR_DICT: dict = {i:i for i in range(3)}


# ---------------------------------------------------------------------------------------------------------------------
def FUNC(*args, **kwargs) -> Any | None:
    pass


def FUNC_NONE(*args, **kwargs) -> None:
    return None


def FUNC_TRUE(*args, **kwargs) -> bool:
    return True


def FUNC_FALSE(*args, **kwargs) -> bool:
    return False


def FUNC_EXC(*args, **kwargs) -> Exception:
    return Exception("FUNC_EXC")


def FUNC_RAISE(*args, _exc: type[BaseException] | None = None, _msg: str | None = None, **kwargs) -> NoReturn:
    _exc = _exc or Exception
    _msg = _msg or "FUNC_RAISE"
    raise _exc(_msg)


def FUNC_ECHO(echo: Any, *args, **kwargs) -> Any:
    return echo


def FUNC_GEN_KEYS(*args, **kwargs) -> Generator:
    """
    NOTE:
        'KEYS' means
            - args - used as keys,
            - if result imply to use only one value from kwargs (like LIST) it will get he exact part KEYS,
        'VALUES' means
            - args - used as values,
            - if result imply to use only one value from kwargs (like LIST) it will get he exact part VALUES,
    """
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    yield from keys


def FUNC_GEN_VALUES(*args, **kwargs) -> Generator:
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    yield from values


def FUNC_ALL_VALUES(*args, **kwargs) -> bool:
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.run as tests
    """
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    return all(values)


def FUNC_ANY_VALUES(*args, **kwargs) -> bool:
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    return any(values)


def FUNC_LIST_KEYS(*args, **kwargs) -> list[Any]:
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    return keys


def FUNC_LIST_VALUES(*args, **kwargs) -> list[Any]:
    keys = [*args, *kwargs]
    values = [*args, *kwargs.values()]
    return values


def FUNC_DICT_KEYS(*args, **kwargs) -> dict[Any, Any | None]:
    result = dict.fromkeys(args)
    result.update(kwargs)
    return result


def FUNC_DICT_VALUES(*args, **kwargs) -> dict[Any, Any | None]:
    result = {f"_k{i}": arg for i, arg in zip(range(len(args)), args)}
    result.update(kwargs)
    return result


# ---------------------------------------------------------------------------------------------------------------------
async def AIO_FUNC(*args, **kwargs) -> Any | None:
    pass


async def AIO_FUNC_NONE(*args, **kwargs) -> None:
    return None


async def AIO_FUNC_TRUE(*args, **kwargs) -> bool:
    return True


async def AIO_FUNC_FALSE(*args, **kwargs) -> bool:
    return False


async def AIO_FUNC_EXC(*args, **kwargs) -> Exception:
    return Exception("AIO_FUNC_EXC")


async def AIO_FUNC_RAISE(*args, **kwargs) -> NoReturn:
    raise Exception("AIO_FUNC_RAISE")


async def AIO_FUNC_ECHO(echo: Any, *args, **kwargs) -> Any:
    return echo


# async def AIO_FUNC_GEN_KEYS(*args, **kwargs) -> Generator:
#     keys = [*args, *kwargs]
#     values = [*args, *kwargs.values()]
#     yield from keys         # Python does not support 'yield from' inside async functions
#
#
# async def AIO_FUNC_GEN_VALUES(*args, **kwargs) -> Generator:
#     keys = [*args, *kwargs]
#     values = [*args, *kwargs.values()]
#     yield from values       # Python does not support 'yield from' inside async functions


# =====================================================================================================================
# AIO_LAMBDA_ARGS: Callable[..., tuple[Any, ...]] = async lambda *args: args        # Expression expected

LAMBDA_ARGS: Callable[..., tuple[Any, ...]] = lambda *args: args        # used as resolve ARGS
LAMBDA_KWARGS: Callable[..., dict[str, Any]] = lambda **kwargs: kwargs  # used as resolve KWARGS
"""
GOAL
----
smth like 
for ARGS - ensure tuple
for KWARGS - use strings without quotes in dicts - but ypu may use directly dict(key1=1, ...)

USAGE
-----
for i in LAMBDA_ARGS(1, 2, 3):
    print(i)
"""


# ---------------------------------------------------------------------------------------------------------------------
LAMBDA: Callable[..., Any | None] = lambda *args, **kwargs: None
LAMBDA_0: Callable[..., int] = lambda *args, **kwargs: 0
LAMBDA_1: Callable[..., int] = lambda *args, **kwargs: 1
LAMBDA_NONE: Callable[..., None] = lambda *args, **kwargs: None
LAMBDA_TRUE: Callable[..., bool] = lambda *args, **kwargs: True
LAMBDA_FALSE: Callable[..., bool] = lambda *args, **kwargs: False

LAMBDA_EXC: Callable[..., Exception] = lambda *args, **kwargs: Exception("LAMBDA_EXC")
# LAMBDA_RAISE = lambda *args, **kwargs: raise Exception("LAMBDA_EXC")      # raise=SyntaxError: invalid syntax
LAMBDA_RAISE: Callable[..., NoReturn] = lambda *args, **kwargs: FUNC_RAISE()
# LAMBDA_GEN_VALUES = lambda *args, **kwargs: yield from range(5)      # yield=SyntaxError: invalid syntax
LAMBDA_GEN_VALUES: Callable[..., Iterable[Any]] = lambda *args, **kwargs: FUNC_GEN_VALUES()
LAMBDA_ECHO: Callable[..., Any] = lambda echo, *args, **kwargs: echo

LAMBDA_ALL_VALUES: Callable[..., bool] = lambda *args, **kwargs: FUNC_ALL_VALUES(*args, **kwargs)
LAMBDA_ANY_VALUES: Callable[..., bool] = lambda *args, **kwargs: FUNC_ANY_VALUES(*args, **kwargs)

LAMBDA_LIST_KEYS = lambda *args, **kwargs: FUNC_LIST_KEYS(*args, **kwargs)
LAMBDA_LIST_VALUES = lambda *args, **kwargs: FUNC_LIST_VALUES(*args, **kwargs)

LAMBDA_DICT_KEYS = lambda *args, **kwargs: FUNC_DICT_KEYS(*args, **kwargs)
LAMBDA_DICT_VALUES = lambda *args, **kwargs: FUNC_DICT_VALUES(*args, **kwargs)


# =====================================================================================================================
class ClsException(Exception):
    pass
INST_EXCEPTION = ClsException("Exception")


# ---------------------------------------------------------------------------------------------------------------------
# class ClsBool(bool):  # cant use it!
#     pass


class ClsInt(int):
    pass


class ClsFloat(float):
    pass


class ClsStr(str):
    pass


class ClsList(list):
    pass


class ClsTuple(tuple):
    pass


class ClsSet(set):
    pass


class ClsDict(dict):
    pass


CLASSES__AS_FUNC: tuple[type, ...] = (ClsInt, ClsFloat, ClsStr, ClsList, ClsTuple, ClsSet, ClsDict, )   # actually this is keep all buildIn


# =====================================================================================================================
class Cls:
    pass
INST = Cls()


class ClsEmpty:
    pass
INST_EMPTY = ClsEmpty()


# ---------------------------------------------------------------------------------------------------------------------
class ClsInitArgsKwargs:
    """
    GOAL
    ----
    just apply init with any args/kwargs
    so no Exception would raise in any case!
    in first idea it was not matter to keep them in instance but did it just in case

    CREATED SPECIALLY FOR
    ---------------------
    ClsBoolTrue/*False
    """
    ARGS: tuple
    KWARGS: dict

    def __init__(self, *args, **kwargs):
        self.ARGS = args
        self.KWARGS = kwargs


class ClsInitRaise:
    def __init__(self, *args, **kwargs) -> NoReturn:
        raise Exception("ClsInitRaise")


# ---------------------------------------------------------------------------------------------------------------------
class ClsCall:
    def __call__(self, *args, **kwargs) -> None:
        pass

    def meth(self, *args, **kwargs) -> None:
        """
        for other results like None/True/False/ClsException use direct LAMBDA/FUNC_*! or wait special necessity.
        """
        pass
INST_CALL = ClsCall()


class ClsCallNone(ClsCall):
    pass
INST_CALL_NONE = ClsCallNone()


class ClsCallTrue:
    def __call__(self, *args, **kwargs) -> bool:
        return True
INST_CALL_TRUE = ClsCallTrue()


class ClsCallFalse:
    def __call__(self, *args, **kwargs) -> bool:
        return False
INST_CALL_FALSE = ClsCallFalse()


class ClsCallRaise:
    def __call__(self, *args, **kwargs) -> NoReturn:
        raise Exception("ClsCallRaise")
INST_CALL_RAISE = ClsCallRaise()


class ClsCallExc:
    def __call__(self, *args, **kwargs) -> Exception:
        return Exception("ClsCallExc")
INST_CALL_EXC = ClsCallExc()


# ---------------------------------------------------------------------------------------------------------------------
class ClsBoolTrue(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        return True
INST_BOOL_TRUE = ClsBoolTrue()


class ClsBoolFalse(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        return False
INST_BOOL_FALSE = ClsBoolFalse()


class ClsBoolRaise(ClsInitArgsKwargs):
    """
    CREATED SPECIALLY FOR
    ---------------------
    funcs.Valid.get_bool as test variant
    """
    def __bool__(self):
        raise Exception()
INST_BOOL_RAISE = ClsBoolRaise()


# ---------------------------------------------------------------------------------------------------------------------
class ClsIterYield:
    """
    CONSTRAINTS
    -----------
    YIELD and RETURN all are acceptable!
    several iterations - work fine!

        class Cls:
        def __iter__(self):
            yield from range(3)
            # return iter(range(3))

        obj = Cls()
        for _ in range(2):
            print()
            for i in obj:
                print(i)
    """

    def __iter__(self):
        # RETURN VARIANTS ------------
        # return [1,2,3]  #TypeError: iter() returned non-iterator of type 'list'
        # return range(5)  #TypeError: iter() returned non-iterator of type 'range'
        # return iter([1,2,3])  #OK

        # YIELD VARIANTS ------------    - MOST PREFERRED!(seek would be reset all time to the beginning!!!)
        # yield from [1,2,3]  #OK
        yield from range(5)   #OK
INST_ITER_YIELD = ClsIterYield()


class ClsIterArgs:
    ARGS: tuple

    def __init__(self, *args):
        self.ARGS = args

    def __iter__(self):
        yield from self.ARGS
INST_ITER_ARGS = ClsIterArgs()


class ClsGen:
    """
    ClsIterNext!
    """
    def __init__(self, start=1, end=3):
        self.start = start
        self.end = end
        self.current = start

    def __iter__(self):
        return self

    def __next__(self):
        if self.current > self.end:
            raise StopIteration
        else:
            result = self.current
            self.current += 1
            return result
INST_GEN = ClsGen()


# ---------------------------------------------------------------------------------------------------------------------
class ClsEq:
    def __init__(self, val: Any = None, *args, **kwargs):
        self.VAL = val

    def __eq__(self, other):
        return other == self.VAL
INST_EQ = ClsEq()


class ClsEqTrue(ClsInitArgsKwargs):
    def __eq__(self, other):
        return True
INST_EQ_TRUE = ClsEqTrue()


class ClsEqFalse(ClsInitArgsKwargs):
    def __eq__(self, other):
        return False
INST_EQ_FALSE = ClsEqFalse()


class ClsEqRaise(ClsInitArgsKwargs):
    def __eq__(self, other):
        raise Exception()
INST_EQ_RAISE = ClsEqRaise()


# ---------------------------------------------------------------------------------------------------------------------
class ClsAw:
    def __await__(self) -> None:
        result = yield from asyncio.sleep(0.1).__await__()
        return result


# ---------------------------------------------------------------------------------------------------------------------
class ClsFullTypes:
    attrZero = 0

    attrNone = None
    attrTrue = True
    attrFalse = False
    attrInt = 1
    attrFloat = 1.1
    attrStr = "str"
    attrBytes = b"bytes"

    attrFunc = FUNC
    attrFuncTrue = FUNC_TRUE
    attrFuncList = FUNC_LIST_KEYS
    attrFuncDict = FUNC_DICT_KEYS
    attrFuncExc = FUNC_EXC
    attrFuncRaise = FUNC_RAISE
    attrFuncGen = FUNC_GEN_VALUES

    attrGenCompr = COMPR_TUPLE_GEN

    attrCls = ClsEmpty
    attrInst = ClsEmpty()
    attrInstMeth = ClsCall().meth

    attrClsCall = ClsCall
    attrInstCall = INST_CALL
    attrClsCallTrue = ClsCallTrue
    attrInstCallTrue = INST_CALL_TRUE
    attrClsCallRaise = ClsCallRaise
    attrInstCallRaise = INST_CALL_RAISE
    attrClsCallExc = ClsCallExc
    attrInstCallExc = INST_CALL_EXC

    attrClsIterYield = ClsIterYield
    attrInstIterYield = INST_ITER_YIELD
    attrClsGen = ClsGen
    attrInstGen = INST_GEN

    attrClsBoolTrue = ClsBoolTrue
    attrInstBoolTrue = INST_BOOL_TRUE
    attrClsBoolFalse = ClsBoolFalse
    attrInstBoolFalse = INST_BOOL_FALSE
    attrClsBoolRaise = ClsBoolRaise
    attrInstBooRaise = INST_BOOL_RAISE

    attrSet = {1,2,3}
    attrList = [1,2,3]
    attrTuple = (1,2,3)
    attrDict = {1:1}
    attrListInst = [*[Cls(), ] * 3, 1]

    @property
    def propertyNone(self) -> None:
        return
    @classmethod
    @property
    def propertyClassmethodNone(cls) -> None:
        return

    @property
    def propertyInt(self) -> int:
        return 1
    @property
    def propertyExc(self) -> Exception:
        return Exception("propertyExc")
    @property
    def propertyRaise(self) -> NoReturn:
        raise Exception("propertyRaise")
    @property
    def propertyFunc(self) -> Callable:
        return FUNC

    def methNone(self) -> None:
        return
    def methInt(self) -> int:
        return 1
    def methExc(self) -> Exception:
        return Exception("methExc")
    def methRaise(self) -> NoReturn:
        raise Exception("methRaise")
    def methFunc(self) -> Callable:
        return FUNC
    @classmethod
    def classmethodNone(cls) -> None:
        return
    @staticmethod
    def staticmethodNone() -> None:
        return
INST_FULL_TYPES = ClsFullTypes()


# ---------------------------------------------------------------------------------------------------------------------
@final
class VALUES_CALLABLE:
    """
    GOAL
    ----
    collect all callables in one place
    """
    LAMBDA: Callable = LAMBDA
    FUNC: Callable = FUNC

    CLS: Callable | type = ClsCall
    INST: Callable = INST_CALL

    METH_CLS: Callable | type = ClsFullTypes.methNone
    METH_CLS_CLASSMETHOD: Callable = ClsFullTypes.classmethodNone
    METH_CLS_STATICMETHOD: Callable = ClsFullTypes.staticmethodNone
    METH_CLS_PROPERTY: Any = ClsFullTypes.propertyNone                           # NOT VALUES_CALLABLE!!!
    METH_CLS_PROPERTY_CLASSMETHOD: Any = ClsFullTypes.propertyClassmethodNone    # NOT VALUES_CALLABLE!!!

    METH_INST: Callable = INST_FULL_TYPES.methNone
    METH_INST_CLASSMETHOD: Callable = INST_FULL_TYPES.classmethodNone
    METH_INST_STATICMETHOD: Callable = INST_FULL_TYPES.staticmethodNone
    METH_INST_PROPERTY: Any = INST_FULL_TYPES.propertyNone                           # NOT VALUES_CALLABLE!!!
    METH_INST_PROPERTY_CLASSMETHOD: Any = INST_FULL_TYPES.propertyClassmethodNone    # NOT VALUES_CALLABLE!!!

    # PROPERTIES is not callables cause of when we access the attribute, decorator will return really final value not callable


def _callable__show_who_really_are():
    # from base_aux.base_types import ObjectInfo
    # ObjectInfo(VALUES_CALLABLE).print()
    # exit()
    from base_aux.aux_attr.m1_annot_attr1_aux import AttrAux_Existed
    for name, item in AttrAux_Existed(VALUES_CALLABLE).dump_dict().items():
        print(f"{name}={callable(item)}")


# ---------------------------------------------------------------------------------------------------------------------
class VictimAttrs:
    # At0
    At1 = None

    An0: Any
    An1: Any = None

    Ct1 = lambda x=None: 1
    Cn1: Callable = lambda x=None: 1


# =====================================================================================================================
if __name__ == "__main__":
    _callable__show_who_really_are()
    VictimAttrs().Cn1()


# =====================================================================================================================



from typing import *
import sys

from base_aux.base_types.m2_info import *
from base_aux.versions.m1_block import *
from base_aux.base_nest_dunders.m2_repr_clsname_str import *
from base_aux.aux_attr.m3_ga1_prefix_1_inst import NestGa_Prefix_RaiseIf


# =====================================================================================================================
def _explore_other():
    # EXPLORE VARIANTS ----------------------------------------------
    # 1=PACKAGING.version ---------
    from packaging import version
    ObjectInfo(version.parse(str((1,2,3)))).print()
    result = version.parse("2.3.1") < version.parse("10.1.2")
    ObjectInfo(version.parse("1.2.3")).print()
    print(result)
    print()

    # 2=PKG_RESOURCES.parse_version ---------
    from pkg_resources import parse_version         # DEPRECATED!!!
    parse_version("1.9.a.dev") == parse_version("1.9a0dev")

    import sys
    print(sys.winver)
    print(sys.version_info)
    print(tuple(sys.version_info))

    result = sys.version_info > (2, 7)
    print(result)


# =====================================================================================================================
TYPE__VERSION_DRAFT = Union[TYPE__VERSION_BLOCK_ELEMENTS_DRAFT,  TYPE__VERSION_BLOCKS_FINAL, 'Version', Any]


# =====================================================================================================================
class Version(NestCmp_GLET_Any, NestRepr__ClsName_SelfStr, NestGa_Prefix_RaiseIf):
    """
    GOAL
    ----
    make a version object from any source
    direct CMP with any Other object

    NOTE
    ----
    VERSION - SPLIT DOTS!
    BLOCK - SPLIT ELEMENTS!

    :ivar SOURCE: try to pass parsed value! it will try to self-parse in _prepare_string, but make it ensured on your own!

    USAGE CMP
    ---------
    Version_Python().raise_if_not__check_ge("2")
    Version_Python().raise_if_not__check_ge("3.11")
    Version_Python().raise_if_not__check_ge("3.11rc1", _comment="need Python GRATER EQUAL")

    FIXME: seems can use EqRaise
    """
    SOURCE: Any = None
    PREPARSE: str = None
    BLOCKS: TYPE__VERSION_BLOCKS_FINAL = ()

    RAISE: bool = True

    def __init__(self, source: Any = None, preparse: str = None, _raise: bool = None) -> None | NoReturn:
        if preparse is not None:
            self.PREPARSE = preparse

        if _raise is not None:
            self.RAISE = _raise

        if source is not None:
            self.SOURCE = source

        self._prepare_source()
        self._parse_blocks()

    # -----------------------------------------------------------------------------------------------------------------
    def _prepare_source(self) -> str:
        """
        ONLY PREPARE STRING FOR CORRECT SPLITTING BLOCKS - parsing blocks would inside VersionBlock
        """

        # TODO: is it need to resolve callables??? - by now NOT!
        # if TypeAux(self.SOURCE).check__callable_func_meth_inst():
        #     value = self.SOURCE()
        # else:
        #     value = self.SOURCE

        if isinstance(self.SOURCE, (list, tuple)):
            result = ".".join([str(block) for block in self.SOURCE])
        else:
            result = str(self.SOURCE)

        result = result.lower()

        # PREPARSE -----
        if self.PREPARSE:
            new = TextAux(result).search__group(self.PREPARSE)
            if new:
                result = new

        # CUT ---------
        for pattern in Pat_Version.VERSION_IN_BRACKETS:
            match = re.search(pattern, result)
            if match:
                result = match[1]
                break

        if "," in result and "." in result and self.RAISE:
            raise Exc__Incompatible_Data(result)
        # else:
        #     result = ""

        for pattern in Pat_Version.VALID_BRACKETS:
            if re.search(pattern, result) and self.RAISE:
                raise Exc__Incompatible_Data(f"{pattern=}/{result=}")

        result = re.sub(r"\A\D+", "", result)   # ver/version
        result = re.sub(r",+", ".", result)
        result = re.sub(r"\.+", ".", result)
        result = result.strip(".")

        self.SOURCE = result
        return self.SOURCE

    def _parse_blocks(self) -> TYPE__VERSION_BLOCKS_FINAL:
        blocks_list__str = str(self.SOURCE).split(".")

        # RESULT -----------
        result = []
        for item in blocks_list__str:
            if not item:
                continue

            try:
                block = VersionBlock(item)
                result.append(block)
            except Exception as exc:
                if self.RAISE:
                    raise Exc__Incompatible_Data(exc)
                else:
                    return ()

        self.BLOCKS = tuple(result)
        return self.BLOCKS

    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self):
        return ".".join([str(block) for block in self.BLOCKS])

    def __bool__(self):
        """
        GOAL
        ----
        if exists at least one NoZERO-block - return True! otherwise return False!
        """
        if len(self) == 0:
            return False
        for block in self.BLOCKS:
            if block != 0:
                return True
        return False
        # else:
        #     return self != "0"

    def __len__(self) -> int:
        return len(self.BLOCKS)

    def __getitem__(self, item: int) -> VersionBlock | None:
        try:
            return self.BLOCKS[item]
        except:
            return

    def __iter__(self):
        yield from self.BLOCKS

    # -----------------------------------------------------------------------------------------------------------------
    @property
    def MAJOR(self) -> VersionBlock | None:
        return self[0]

    @property
    def MINOR(self) -> VersionBlock | None:
        return self[1]

    @property
    def MICRO(self) -> VersionBlock | None:
        return self[2]

    # -----------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other: TYPE__VERSION_DRAFT) -> int | NoReturn:
        if not bool(self) and not bool(other):
            return 0

        other = self.__class__(other, _raise=self.RAISE)

        if not bool(self) and not bool(other):
            return 0

        # equel ----------------------
        if str(self) == str(other):
            return 0

        # by elements ----------------
        for block_1, block_2 in zip(self, other):
            if block_1 == block_2:
                continue
            else:
                return int(block_1 > block_2) or -1

        # final - longest ------------
        return int(len(self) > len(other)) or -1

    # -----------------------------------------------------------------------------------------------------------------
    raise_if__check_eq: Callable[..., NoReturn | None]
    raise_if_not__check_eq: Callable[..., NoReturn | None]

    raise_if__check_ne: Callable[..., NoReturn | None]
    raise_if_not__check_ne: Callable[..., NoReturn | None]

    raise_if__check_le: Callable[..., NoReturn | None]
    raise_if_not__check_le: Callable[..., NoReturn | None]

    raise_if__check_lt: Callable[..., NoReturn | None]
    raise_if_not__check_lt: Callable[..., NoReturn | None]

    raise_if__check_ge: Callable[..., NoReturn | None]
    raise_if_not__check_ge: Callable[..., NoReturn | None]

    raise_if__check_gt: Callable[..., NoReturn | None]
    raise_if_not__check_gt: Callable[..., NoReturn | None]


# =====================================================================================================================



# TODO: apply aio??? is it possible??? think no!

import time
import pytest

# from base_aux.base_types import TypeAux   # CIRCULAR IMPORT

from base_aux.base_enums.m2_enum1_adj import *
from base_aux.base_nest_dunders.m1_init1_source2_kwargs import *
from base_aux.base_nest_dunders.m3_calls import *


# =====================================================================================================================
# class LambdaSimple(NestInit_SourceKwArgs_Implicit, NestCall_Resolve):
#     """
#     NOTE
#     ----
#     CANT USE IT AS OBJECT IN THREAD!!!
#     so just use Lambda
#
#     GOAL
#     ----
#     simple replace common lambda func!
#     """
#     # SOURCE: Union[Callable[..., Any], Any, type]
#     #
#     # # =================================================================================================================
#     # def resolve(self, *args, **kwargs) -> Any | NoReturn:
#     #     if callable(self.SOURCE):
#     #         result = self.SOURCE(*args, **kwargs)
#     #     else:
#     #         result = self.SOURCE
#     #     return result
#     #
#     # # def __eq__(self, other: Any) -> bool | NoReturn:      # NOTE: DONT USE EQ!
#     # #     return EqAux(self()).check_doubleside__bool(other)
#     #
#     # def __bool__(self) -> bool | NoReturn:
#     #     return bool(self())
#     #
#     # # =================================================================================================================
#     # def check_raise(self, *args, **kwargs) -> bool:
#     #     try:
#     #         result = self(*args, **kwargs)
#     #         return False
#     #     except:
#     #         return True
#     #
#     # def check_no_raise(self, *args, **kwargs) -> bool:
#     #     return not self.check_raise(*args, **kwargs)
#     #
#     # def wait_finished(self, sleep: float = 1) -> None:
#     #     """
#     #     GOAL
#     #     ----
#     #     run if not started yet
#     #     then wait finished
#     #     """
#     #     if self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.NONE:
#     #         self.run()
#     #
#     #     count = 1
#     #     while self.PROCESS_ACTIVE != EnumAdj_ProcessStateActive.FINISHED:
#     #         print(f"wait_finished {count=}")
#     #         count += 1
#     #         time.sleep(sleep)


# =====================================================================================================================
class Lambda(NestInit_SourceKwArgs_Implicit, NestCall_Resolve):
    """
    GOAL
    ----
    1. (MAIN) delay probable raising on direct func execution (used with NestInit_AttrsLambdaResolve)
    like creating base_types on Cls attributes
        class Cls:
            ATTR = PrivateValues(123)   # -> Lambda(PrivateValues, 123) - IT IS OLD!!!! but could be used as example!

    2. (not serious) replace simple lambda!
    by using lambda you should define args/kwargs any time! and im sick of it!
        func = lambda *args, **kwargs: sum(*args) + sum(**kwargs.values())  # its not a simple lambda!
        func = lambda *args: sum(*args)  # its simple lambda
        result = func(1, 2)
    replace to
        func = Lambda(sum)
        result = func(1, 2)

        func = Lambda(sum, 1, 2)
        result = func()
    its те a good idea to replace lambda fully!
    cause you cant replace following examples
        func_link = lambda source: str(self.Victim(source))
        func_link = lambda source1, source2: self.Victim(source1) == source2

    NOTE
    ----
    no calling on init!

    SPECIALLY CREATED FOR
    ---------------------
    Item for using with NestInit_AttrsLambdaResolve

    WHY NOT 1=simple LAMBDA?
    ------------------------
    extremely good point!
    but
    1. in case of at least NestInit_AttrsLambdaResolve you cant distinguish method or callable attribute!
    so you explicitly define attributes/base_types for later constructions
    and in some point it can be more clear REPLACE LAMBDA by this solvation!!!

    WHY NOT 2=CallableAux
    ------------------------
    here is not intended using indirect result like Exception! just raise if raised! so not safe state!

    NOTE
    ----
    CANT REPLACE LAMBDA IN ANY CASE!
        func_link = lambda *_args: getattr(victim, meth)(*_args)
    - will call at same time by Lambda, and if meth is not exists - return:
        Lambda(getattr(victim, meth), *args)

    TIP
    ---
    if need ARGS resolve by SingleMulty - do it before
        self.ARGS = ArgsKwargsAux(args).resolve_args()

    """
    SOURCE: Union[Callable[..., Any], Any, type]

    # thread ready -----
    PROCESS_ACTIVE: EnumAdj_ProcessStateActive = EnumAdj_ProcessStateActive.NONE
    RESULT: Any = None
    EXC: BaseException | None = None

    # UNIVERSAL =======================================================================================================
    def run(self, *args, **kwargs) -> None:
        """
        NOTE
        ----
        DONT USE for getting result! only to execute calculation process!!!
        and thread ready for start usage!
        """
        # ONLY ONE EXECUTION on instance!!! dont use locks! -------------
        if self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.STARTED:
            return

        # WORK ----------------------------------------------------------
        self.PROCESS_ACTIVE = EnumAdj_ProcessStateActive.STARTED
        self.RESULT = None
        self.EXC = None

        args = args or self.ARGS
        kwargs = {**self.KWARGS, **kwargs}

        try:
            # if self.SOURCE == NoValue:
            #     self.RESULT = self.SOURCE
            if callable(self.SOURCE):  # callable accept all variants! TypeAux.check__callable_func_meth_inst_cls!
                self.RESULT = self.SOURCE(*args, **kwargs)
            else:
                self.RESULT = self.SOURCE
        except BaseException as exc:
            print(f"{exc!r}")
            self.EXC = exc

        # FIN ----------------------------------------------------------
        self.PROCESS_ACTIVE = EnumAdj_ProcessStateActive.FINISHED

    def wait_finished(self, sleep: float = 1, run: bool = None) -> None:
        """
        GOAL
        ----
        run if not started yet
        then wait finished

        NOTE
        ----
        dont forget using always after run!!!
        """
        if run or self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.NONE:
            self.run()

        count = 1
        while self.PROCESS_ACTIVE != EnumAdj_ProcessStateActive.FINISHED:
            print(f"wait_finished {count=}")
            count += 1
            time.sleep(sleep)

    # =================================================================================================================
    def resolve(self, *args, **kwargs) -> Any | NoReturn:
        # OVERWRITE for derivatives!!!!
        self.run(*args, **kwargs)
        self.wait_finished()

        # FIN ----------------------------------------------------------
        if self.EXC is not None:
            raise self.EXC
        else:
            return self.RESULT

    def __eq__(self, other: Any) -> bool | NoReturn:        # TODO: decide deprecate???
        return self() == other

    def __bool__(self) -> bool | NoReturn:
        return bool(self())

    # =================================================================================================================
    def check_raised__bool(self, _EXPECTED: bool = True) -> bool:
        self.run()
        self.wait_finished()

        # FIN ----------------------------------------------------------
        if self.EXC is not None:
            return _EXPECTED is True
        else:
            return _EXPECTED is False

    def check_no_raised__bool(self, _EXPECTED: bool = True) -> bool:
        self.run()
        self.wait_finished()

        # FIN ----------------------------------------------------------
        if self.EXC is not None:
            return _EXPECTED is False
        else:
            return _EXPECTED is True

    # ----------------------------------------------------------------
    def check_raised__assert(self, _EXPECTED: bool = True) -> None | NoReturn:
        assert self.check_raised__bool() is _EXPECTED

    def check_no_raised__assert(self, _EXPECTED: bool = True) -> None | NoReturn:
        assert self.check_no_raised__bool() is _EXPECTED

    # =================================================================================================================
    def check_expected__assert(
            self,
            # args: TYPING.ARGS_DRAFT = (),      # DONT USE HERE!!!
            # kwargs: TYPING.KWARGS_DRAFT = None,

            _EXPECTED: TYPING.EXPECTED = True,
            # EXACT VALUE (noCallable) OR AnyCLass - to cmp as isinstanceOrSubclass!!!
            _MARK: pytest.MarkDecorator | None = None,
            _COMMENT: str | None = None,
    ) -> None | NoReturn:
        """
        NOTE
        ----
        this is same as funcs.Valid! except following:
            - if validation is Fail - raise assert!
            - no skips/cumulates/logs/ last_results/*values

        GOAL
        ----
        test target func/obj with exact parameters
        no exception withing target func!

        SPECIALLY CREATED FOR
        ---------------------
        unit tests by pytest
        """
        args = self.ARGS
        kwargs = self.KWARGS
        comment = _COMMENT or ""

        # MARKS -------------------------
        # print(f"{pytest.mark.skipif(True)=}")
        if _MARK == pytest.mark.skip:
            pytest.skip("skip")
        elif isinstance(_MARK, pytest.MarkDecorator) and _MARK.name == "skipif" and all(_MARK.args):
            pytest.skip("skipIF")

        try:
            actual_value = self.resolve(*args, **kwargs)
        except Exception as exc:
            actual_value = exc  # this is an internal value! when use incorrect ArgsKw!!!

        try:
            print(f"Expected[{self.SOURCE}/{args=}/{kwargs=}//{actual_value=}/{_EXPECTED=}]")
        except Exception as exc:
            print(f"!iternal exc! {self.__class__.__name__}({exc=})")
            for part_name, part_obj in dict(
                    source=self.SOURCE,
                    args=args,
                    kwargs=kwargs,
                    actual_value=actual_value,
                    _EXPECTED=_EXPECTED,
            ).items():
                try:
                    print(f"{part_name=}/{part_obj=}")
                except Exception as exc:
                    print(f"{part_name=}/{exc=}")

        result = TypeAux(actual_value).check__subclassed_or_isinst__from_cls(_EXPECTED)

        if not result:
            # result = EqAux(actual_value).check_doubleside__bool(_EXPECTED)

            try:
                result = actual_value == _EXPECTED
            except:
                pass

        if not result:
            # result = EqAux(actual_value).check_doubleside__bool(_EXPECTED)

            try:
                result = _EXPECTED == actual_value
            except:
                pass

        if _MARK == pytest.mark.xfail:
            assert not result, f"[xfail]{comment}"
        else:
            assert result

    def check_expected__bool(self, *args, **kwargs) -> bool:
        """
        GOAL
        ----
        extend work for not only in unittests
        """
        try:
            self.check_expected__assert(*args, **kwargs)
            return True
        except:
            return False

    # =================================================================================================================
    def resolve__style(self, callable_use: EnumAdj_CallResolveStyle = EnumAdj_CallResolveStyle.RAISE, *args, **kwargs) -> Any | None | Exception | NoReturn | EnumAdj_CallResolveStyle | bool:
        """
        NOTE
        ----
        it is just a collection for all variants in one func!

        it is not so convenient to use param callable_use!
        SO preferred using other/further direct methods!
        """
        if callable_use == EnumAdj_CallResolveStyle.DIRECT:
            return self.SOURCE

        elif callable_use == EnumAdj_CallResolveStyle.EXC:
            return self.resolve__exc(*args, **kwargs)

        elif callable_use == EnumAdj_CallResolveStyle.RAISE:
            return self.resolve__raise(*args, **kwargs)

        elif callable_use == EnumAdj_CallResolveStyle.RAISE_AS_NONE:
            return self.resolve__raise_as_none(*args, **kwargs)

        elif callable_use == EnumAdj_CallResolveStyle.SKIP_CALLABLE:
            return self.resolve__skip_callables(*args, **kwargs)

        elif callable_use == EnumAdj_CallResolveStyle.SKIP_RAISED:
            return self.resolve__skip_raised(*args, **kwargs)

        elif callable_use == EnumAdj_CallResolveStyle.BOOL:
            return self.resolve__bool(*args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def resolve__exc(self, *args, **kwargs) -> Any | Exception:
        """
        GOAL
        ----
        same as resolve_raise but
        attempt to simplify result by not using try-sentence.
        so if get raise in resolve_raise - return ClsException object

        USEFUL IDEA
        -----------
        1. in gui when its enough to get str() on result and see the result

        SPECIALLY CREATED FOR
        ---------------------
        just in case

        """
        try:
            return self(*args, **kwargs)
        except Exception as exc:
            return exc

    def resolve__raise(self, *args, **kwargs) -> Any | NoReturn:
        """
        just a direct result for call

        SPECIFIC LOGIC
        --------------
        if callable - call and return result.
        else - return source.

        GOAL
        ----
        get common expected for any python code result - simple calculate or raise!
        because of resolve_exc is not enough!

        SPECIALLY CREATED FOR
        ---------------------
        NestGa_Prefix
        check Privates in pytest for skipping

        USE Lambda_TrySuccess instead!
        """
        return self(*args, **kwargs)

    def resolve__raise_as_none(self, *args, **kwargs) -> Any | None:
        try:
            return self.resolve__raise(*args, **kwargs)
        except:
            return None

    def resolve__skip_callables(self, *args, **kwargs) -> Any | NoReturn:
        if callable(self.SOURCE):
            return VALUE_SPECIAL.SKIPPED
        else:
            return self.SOURCE

    def resolve__skip_raised(self, *args, **kwargs) -> Any | NoReturn:
        try:
            return self.resolve__raise(*args, **kwargs)
        except:
            return VALUE_SPECIAL.SKIPPED

    def resolve__bool(self, *args, **kwargs) -> bool:
        """
        GOAL
        ----
        same as resolve_exc but
        apply bool func on result

        ability to get bool result with meanings:
            - methods/funcs must be called
                assert get_bool(LAMBDA_TRUE) is True
                assert get_bool(LAMBDA_NONE) is False

            - Exceptions assumed as False
                assert get_bool(Exception) is False
                assert get_bool(Exception("FAIL")) is False
                assert get_bool(LAMBDA_EXC) is False

            - for other values get classic bool()
                assert get_bool(None) is False
                assert get_bool([]) is False
                assert get_bool([None, ]) is True

                assert get_bool(LAMBDA_LIST) is False
                assert get_bool(LAMBDA_LIST, [1, ]) is True

            - if on bool() exception raised - return False!
                assert get_bool(ClsBoolRaise()) is False

        CREATED SPECIALLY FOR
        ---------------------
        funcs.Valid.skip_link or else value/func assumed as bool result
        """
        try:
            result = self.resolve__raise(*args, **kwargs)
            try:
                is_exc = issubclass(result, Exception)  # keep first
            except:
                is_exc = isinstance(result, Exception)

            if is_exc:
                return False
            return bool(result)
        except:
            return False


# =====================================================================================================================



import pathlib
import re
import sys
from typing import *

from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.base_types.m1_type_aux import TypeAux
from base_aux.cli.m1_cli_user import CliUser
from base_aux.versions.m2_version import Version


# =====================================================================================================================
# TODO: FINISH!!!
# TODO: FINISH!!!
# TODO: FINISH!!!
# TODO: FINISH!!!
# TODO: FINISH!!!
# TODO: FINISH!!!


# =====================================================================================================================
class PyModule:
    TARGET_NAME: str
    TARGET_VER: str = None

    def __init__(self, name: str):
        # if
        # self.NAME = name
        pass


PATTERN_IMPORT__MULTY_COMMA = r"(?:^|\n) *import +(\w+(?:\.\w+)* *(?:, *\w+(?:\.\w+)* *)*) *"
PATTERN_IMPORT__MULTY_COMMA_BRACKETS = r"(?:^|\n) *import +\( *(\w+(?:\.\w+)* *(?:,\s*\w+(?:\.\w+)* *)*) *\) *"
PATTERN_IMPORT__FROM = r"(?:^|\n) *from +(\w+(?:\.\w+)*) +import +"
PATTERNS_IMPORT = [PATTERN_IMPORT__MULTY_COMMA, PATTERN_IMPORT__MULTY_COMMA_BRACKETS, PATTERN_IMPORT__FROM]


class CmdPattern:
    """
    USAGE
    -----
    cmd = pattern % (PY_PATH, PARAMETER)
    cmd = CmdPattern.INSTALL_UPGRADE__MODULE % (self.PY_PATH, modules)
    """
    INSTALL__MODULE: str = f"%s -m pip install %s"    # (PY_PATH, MODULE)
    INSTALL__FILE: str = f"%s -m pip install -r %s"   # (PY_PATH, FILE)

    INSTALL_UPGRADE__MODULE: str = f"%s -m pip install --upgrade %s"    # (PY_PATH, MODULE)
    INSTALL_UPGRADE__FILE: str = f"%s -m pip install --upgrade -r %s"   # (PY_PATH, FILE)

    SHOW_INFO: str = f"%s -m pip show %s"                               # (PY_PATH, MODULE)
    UNINSTALL: str = f"%s -m pip uninstall -y %s"                       # (PY_PATH, MODULE)


# =====================================================================================================================
class Packages:
    """
    RULE
    ----
    for all module names you could use
    """
    PKGSET__PyPI_DISTRIBUTION: list[str] = [
        # =============================
        # DISTRIBUTION PyPI
        # -----------------------------
        # new setup
        "build",
        "twine",        # necessary! share project on pypi!
        "setuptools",   # necessary! if not

        # old setup (maybe not all)
        "sdist",                # not necessary! works without it on new!
        "bdist-wheel-name",     # not necessary! works without it on new!

    ]
    PY_PATH: str = sys.executable
    cli: CliUser

    # FILENAME_457: str = "requirements__centoid457.txt"      # FILE WILL NOT GO WITHING MODULE AS PART!
    # FILEPATH_457: pathlib.Path = pathlib.Path(__file__).parent.joinpath(FILENAME_457)

    # def explore(self):
    #     # print(f"{__file__=}")
    #     # print(f"{pathlib.Path(__file__).parent.joinpath(self.FILENAME_457)=}")
    #     # print(f"{pathlib.Path(__file__).parent.joinpath(self.FILENAME_457).exists()=}")
    #
    #     print(self.FILEPATH_457)
    #     print(f"{self.FILEPATH_457.exists()}")

    # TODO: add InitSource for moduleNames!
    def __init__(self, py_path: str = None):
        if py_path is not None:
            self.PY_PATH = py_path

        self.cli = CliUser()

    # =================================================================================================================
    def install(self, modules: Union[str, list[str]]) -> bool:
        # LIST -----------------------------------------------
        if isinstance(modules, (list, tuple, set, )):
            result = all([self.install(module) for module in modules])
            return result

        # SINGLE -----------------------------------------------
        cmd = CmdPattern.INSTALL__MODULE % (self.PY_PATH, modules)
        self.cli.send(cmd, timeout=60 * 2, print_all_states=False)

        return self.cli.last_finished_success

    def reinstall(self, module: str) -> None:
        """
        # FIXME: NOT TESTES!

        GOAL
        ----
        its not just an idea for the sake of just an idea!
        sometimes (and quite often in my experience) i got error when just updating the module!
        """

        self.uninstall(module)
        self.install(module)

    def upgrade(self, modules: Union[str, list[str]]) -> bool:
        """
        you can use explicit version definitions!
            name==0.0.1

NOT EXISTS
==================================================
[CLI_SEND] [python -m pip install --upgrade singleton-meta]
==================================================
self.counter=1
self.counter_in_list=0
self.last_cmd='python -m pip install --upgrade singleton-meta'
self.last_duration=1.347492
self.last_finished=True
self.last_finished_success=True
self.last_retcode=0
--------------------------------------------------
self.last_stdout=
	|'Collecting singleton-meta'
	|'  Using cached singleton_meta-0.1.1-py3-none-any.whl.metadata (2.8 kB)'
	|'Using cached singleton_meta-0.1.1-py3-none-any.whl (5.4 kB)'
	|'Installing collected packages: singleton-meta'
	|'Successfully installed singleton-meta-0.1.1'
	|''
--------------------------------------------------
self.last_stderr=
--------------------------------------------------
self.last_exc_timeout=None
==================================================



ALREADY OK
==================================================
[CLI_SEND] [python -m pip install --upgrade singleton-meta]
==================================================
self.counter=1
self.counter_in_list=0
self.last_cmd='python -m pip install --upgrade singleton-meta'
self.last_duration=1.39782
self.last_finished=True
self.last_finished_success=True
self.last_retcode=0
--------------------------------------------------
self.last_stdout=
	|'Requirement already satisfied: singleton-meta in c:\\python3.12.0x64\\lib\\site-packages (0.1.1)'
	|''
--------------------------------------------------
self.last_stderr=
--------------------------------------------------
self.last_exc_timeout=None
==================================================



GOOD UPGRADE
==================================================
[CLI_SEND] [python -m pip install --upgrade requirements-checker]
==================================================
self.counter=1
self.counter_in_list=0
self.last_cmd='python -m pip install --upgrade requirements-checker'
self.last_duration=1.367846
self.last_finished=True
self.last_finished_success=True
self.last_retcode=0
--------------------------------------------------
self.last_stdout=
	|'Requirement already satisfied: requirements-checker in c:\\!=starichenko=element\\!=projects\\abc=requirements (0.0.7)'
	|'Collecting requirements-checker'
	|'  Using cached requirements-0.1.0-py3-none-any.whl.metadata (2.2 kB)'
	|'Using cached requirements-0.1.0-py3-none-any.whl (7.8 kB)'
	|'Installing collected packages: requirements-checker'
	|'  Attempting uninstall: requirements-checker'
	|'    Found existing installation: requirements-checker 0.0.7'
	|"    Can't uninstall 'requirements-checker'. No files were found to uninstall."
	|'Successfully installed requirements-checker-0.1.0'
	|''
--------------------------------------------------
self.last_stderr=
--------------------------------------------------
self.last_exc_timeout=None
==================================================



FIXME: sometimes modules have incorrect SHARE!!! maybe need check upgrade after installation!!! and show ERROR!!
ERROR ON DISTRIBUTION
==================================================
[#####################ERROR#####################]
self.counter=1
self.counter_in_list=0
self.last_cmd='python -m pip install --upgrade privates'
self.last_duration=1.392156
self.last_finished=True
self.last_finished_success=False
self.last_retcode=1
--------------------------------------------------
self.last_stdout=
        |'Collecting privates'
        |'  Using cached privates-0.5.4.tar.gz (8.8 kB)'
        |'  Preparing metadata (setup.py): started'
        |"  Preparing metadata (setup.py): finished with status 'error'"
        |''
--------------------------------------------------
self.last_stderr=
        |'  error: subprocess-exited-with-error'
        |'  '
        |'  python setup.py egg_info did not run successfully.'
        |'  exit code: 1'
        |'  '
        |'  [6 lines of output]'
        |'  Traceback (most recent call last):'
        |'    File "<string>", line 2, in <module>'
        |'    File "<pip-setuptools-caller>", line 34, in <module>'
        |'    File "C:\\Users\\a.starichenko\\AppData\\Local\\Temp\\pip-install-u6219r9f\\private-values_f5cf3965eb014632a70ff97372b73571\\setup.py", line 2, in <module>'
        |'      from PROJECT import PROJECT'
        |"  ModuleNotFoundError: No module named 'PROJECT'"
        |'  [end of output]'
        |'  '
        |'  note: This error originates from a subprocess, and is likely not a problem with pip.'
        |'error: metadata-generation-failed'
        |''
        |'Encountered error while generating package metadata.'
        |''
        |'See above for output.'
        |''
        |'note: This is an issue with the package mentioned above, not pip.'
        |'hint: See above for details.'
        |''
--------------------------------------------------
self.last_exc_timeout=None
==================================================
"""
        # LIST -----------------------------------------------
        if isinstance(modules, (list, tuple, set, )):
            result = all([self.upgrade(module) for module in modules])
            return result

        # SINGLE -----------------------------------------------
        cmd = CmdPattern.INSTALL_UPGRADE__MODULE % (self.PY_PATH, modules)
        self.cli.send(cmd, timeout=60 * 2, print_all_states=False)

        # TEXT_CUM ===================================================
        text_cum = f"[{modules}]"
        # RESULT EXISTS ---------------------------------------------
        while True:
            # result EXISTS OLD ---------------------------------------------   # KEEP IT FIRST!!!
            # 'Found existing installation: requirements-checker 0.0.7'
            match = re.search(r'Found existing installation: \S+ (\d+\.\d+\.\d+)\s', self.cli.last_stdout)
            if match:
                text_cum += f"(existed old){match[1]}"
                # -----THIS IDEA IS NOT WORK!!!
                # FIXME: TODO: need check version before send UpdateCmd! then compare and decide to reinstall!!!!
                # self.uninstall(modules)     # need uninstall! for sure!
                # self.install(modules)

                """
[CLI_SEND] [('python -m build --sdist -n', 60)]
.........
==================================================
[#####################ERROR#####################]
self.counter=3
self.counter_in_list=1
self.last_cmd='python -m build --sdist -n'
self.last_duration=0.593
self.last_finished=True
self.last_finished_success=False
self.last_retcode=1
--------------------------------------------------
self.last_stdout=
        |'* Getting build dependencies for sdist...'
        |'running egg_info'
        |'writing buses.egg-info\\PKG-INFO'
        |'writing dependency_links to buses.egg-info\\dependency_links.txt'
        |'writing top-level names to buses.egg-info\\top_level.txt'
        |"reading manifest file 'buses.egg-info\\SOURCES.txt'"
        |"adding license file 'LICENSE'"
        |''
        |'ERROR Backend subprocess exited when trying to invoke get_requires_for_build_sdist'
        |''
--------------------------------------------------
self.last_stderr=
        |'Traceback (most recent call last):'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\pyproject_hooks\\_in_process\\_in_process.py", line 353, in <module>'
        |'    main()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\pyproject_hooks\\_in_process\\_in_process.py", line 335, in main'
        |"    json_out['return_val'] = hook(**hook_input['kwargs'])"
        |'                             ^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\pyproject_hooks\\_in_process\\_in_process.py", line 287, in get_requires_for_build_sdist'
        |'    return hook(config_settings)'
        |'           ^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\build_meta.py", line 330, in get_requires_for_build_sdist'
        |'    return self._get_build_requires(config_settings, requirements=[])'
        |'           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\build_meta.py", line 297, in _get_build_requires'
        |'    self.run_setup()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\build_meta.py", line 497, in run_setup'
        |'    super().run_setup(setup_script=setup_script)'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\build_meta.py", line 313, in run_setup'
        |'    exec(code, locals())'
        |'  File "<string>", line 22, in <module>'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\__init__.py", line 108, in setup'
        |'    return distutils.core.setup(**aux_attr)'
        |'           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_distutils\\core.py", line 184, in setup'
        |'    return run_commands(dist)'
        |'           ^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_distutils\\core.py", line 200, in run_commands'
        |'    dist.run_commands()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_distutils\\dist.py", line 970, in run_commands'
        |'    self.run_command(cmd)'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\dist.py", line 945, in run_command'
        |'    super().run_command(command)'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_distutils\\dist.py", line 989, in run_command'
        |'    cmd_obj.run()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\command\\egg_info.py", line 310, in run'
        |'    self.find_sources()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\command\\egg_info.py", line 318, in find_sources'
        |'    mm.run()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\command\\egg_info.py", line 544, in run'
        |'    self.prune_file_list()'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\command\\egg_info.py", line 610, in prune_file_list'
        |'    base_dir = self.distribution.get_fullname()'
        |'               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_core_metadata.py", line 266, in get_fullname'
        |'    return _distribution_fullname(self.get_name(), self.get_version())'
        |'           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |'  File "C:\\Python3117x64\\Lib\\site-packages\\setuptools\\_core_metadata.py", line 284, in _distribution_fullname'
        |'    canonicalize_version(version, strip_trailing_zero=False),'
        |'    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'
        |"TypeError: canonicalize_version() got an unexpected keyword argument 'strip_trailing_zero'"
        |''
--------------------------------------------------
self.last_exc_timeout=None
==================================================
[ERROR] cmd_item=('python -m build --sdist -n', 60) in full sequence cmd=[('python -m build --sdist -n', 60), ('python -m build --wheel -n', 60), ('twine upload dist/*', 90)]



[FINISHED] (result=False) - press Enter to close
[FINISHED] (result=False) - press Enter to close
[FINISHED] (result=False) - press Enter to close
[FINISHED] (result=False) - press Enter to close
[FINISHED] (result=False) - press Enter to close
[FINISHED] (result=False) - press Enter to close
                """
                # -----THIS IDEA IS NOT WORK!!!
                break

            # 'Requirement already satisfied: requirements-checker in c:\\!=starichenko=element\\!=projects\\abc=requirements (0.0.7)'
            match = re.search(r'Requirement already satisfied: .+ \((\d+\.\d+\.\d+)\)', self.cli.last_stdout)
            if len(self.cli.last_stdout.split("\n")) == 2 and match:
                text_cum += f"(already new){match[1]}"
                break

            # result NEW VERSION! ---------------------------------------------
            # 'Successfully installed singleton-meta-0.1.1'
            match = re.search(r'Successfully installed \S+-(\d+\.\d+\.\d+)\s', self.cli.last_stdout)
            if match:
                text_cum += f"->{match[1]}(upgraded new)"
                break

            # FINAL STOP
            break

        # FINISH ===================================================
        print(text_cum)
        return self.cli.last_finished_success

    def upgrade_pip(self) -> bool:
        """
        upgrade PIP module
        """
        return self.upgrade("pip")

    def upgrade_prj(self, project: "PROJECT") -> bool:
        """
        upgrade PROJECT module to last

        CREATED specially for ensure upgrade to last version in active GIT-repo (you need to start in from git-repo!)
        """
        while not self.check_prj_installed_latest(project):
            self.upgrade(project.NAME_INSTALL)

        return True

    # =================================================================================================================
    def check_prj_installed_latest(self, project: "PROJECT") -> bool:
        """
        check version is actual latest!
        if prj ver is not same as installed - return False

        CREATED SPECIALLY FOR
        ---------------------
        check prj version before publish in pypi
        """
        prj_name = project.NAME_INSTALL
        ver_prj = project.VERSION

        ver_active = self.version_get_installed(prj_name)
        if not ver_active:
            return False

        print(f"{ver_prj=}/{ver_active=}")
        return Version(ver_active) == ver_prj

    # =================================================================================================================
    def upgrade_file(self, filepath: Union[str, pathlib.Path], bylines: bool = None) -> bool:
        """
        upgrade modules by file requirements.py

        :param bylines: not applied yet! # TODO: supposed to install over direct lines
        """
        filepath = pathlib.Path(filepath)
        if not filepath.exists():
            msg = f"[ERROR] file not found {filepath}"
            print(msg)
            return False

        cmd = CmdPattern.INSTALL_UPGRADE__FILE % (self.PY_PATH, filepath)
        print("-" *20)
        print(f"{filepath=}")
        print(filepath.read_text(encoding="utf8"))
        print("-" *20)
        return self.cli.send(cmd, timeout=60 * 5)

    def check_file(self, file) -> bool:
        """
        # TODO: CREATE
        """
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        pass

    # =================================================================================================================
    def version_get_pypi(self, name: str) -> Optional[str]:
        """
        # TODO: CREATE
        """
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        pass

    def version_get_installed(self, name: str) -> Optional[str]:
        """
        get version for module if it installed.

        C:\\Users\\a.starichenko>pip show object-info
        Name: object-info
        Version: 0.1.12
        Summary: print info about object (attributes+properties+methods results)
        Home-page: https://github.com/centroid457/
        Author: Andrei Starichenko
        Author-email: centroid@mail.ru
        License:
        Location: C:\\Python3.12.0x64\\Lib\\site-packages
        Requires:
        Required-by:

        C:\\Users\\a.starichenko>pip show base_types
        Name: object-info
        Version: 0.1.12
        Summary: print info about object (attributes+properties+methods results)
        Home-page: https://github.com/centroid457/
        Author: Andrei Starichenko
        Author-email: centroid@mail.ru
        License:
        Location: C:\\Python3.12.0x64\\Lib\\site-packages
        Requires:
        Required-by:

        C:\\Users\\a.starichenko>
        """
        cmd = CmdPattern.SHOW_INFO % (self.PY_PATH, name)
        if self.cli.send(cmd, timeout=2, print_all_states=False) and self.cli.last_stdout:
            match = re.search(r'Version: (\d+\.\d+\.\d+)\s*', self.cli.last_stdout)
            if match:
                return match[1]

    def check_installed(self, modules: Union[str, list[str]]) -> bool:
        # LIST -----------------------------------------------
        if isinstance(modules, (list, tuple, set, )):
            for module in modules:
                if not self.check_installed(module):
                    return False
            return True

        # ONE -----------------------------------------------
        return self.version_get_installed(modules) is not None

    def uninstall(self, modules: Union[str, list[str]]) -> None:
        # LIST -----------------------------------------------
        if isinstance(modules, (list, tuple, set, )):
            for module in modules:
                self.uninstall(module)

        # ONE -----------------------------------------------
        if modules in self.PKGSET__PyPI_DISTRIBUTION:
            return

        cmd = CmdPattern.UNINSTALL % (self.PY_PATH, modules)
        self.cli.send(cmd, timeout=120)

    # =================================================================================================================
    @staticmethod
    def parse__requirements(filepath: AnyStr | pathlib.Path) -> list[str]:
        """
        GOAL
        ----
        get req modules from requirements.txt
        so just parse a req file

        SPECIALLY CREATED FOR
        ---------------------
        setup.py - get values for install_requires param!
        """
        filepath = pathlib.Path(filepath)
        if not filepath.exists():
            msg = f"[ERROR] file not found {filepath}"
            print(msg)
            return []

        text = filepath.read_text(encoding="utf8")
        print("-" * 20)
        print(f"{filepath=}")
        print(f"{text=}")
        print("-" * 20)

        result = TextAux(text).parse__requirements_lines()

        return result

    @staticmethod
    def parse_import_modules__txt(text: AnyStr, patterns: list[AnyStr] | AnyStr) -> list[str]:
        """
        GOAL
        ----
        get imported modules from text (python sourcecode)
        get only root names (no parts like pkg.module)
        """
        # SINGLE -------------
        if not TypeAux(patterns).check__iterable_not_str():
            patterns = [patterns, ]

        # ITERABLE -------------
        result = []
        for pattern in patterns:
            items = re.findall(pattern, text)
            for modules_txt in items:
                modules_txt = re.sub(r"\s*", "", modules_txt)
                modules_list = modules_txt.split(",")
                for module_new in modules_list:
                    module_path = module_new.split(".")
                    module_root = module_path[0]
                    if module_root not in result:
                        result.append(module_root)
        return result

    @classmethod
    def find_in_files(
            cls,
            patterns: list[AnyStr] | AnyStr,
            path: Union[str, pathlib.Path] | None = None,
            print_empty_files: bool | None = None,
            skip_paths: Union[str, list[str]] | None = None,
            glob_mask: str = '**/*.py',
    ) -> list[str]:
        """
        GOAL
        ----
        find values by patterns in files
        """
        if not TypeAux(patterns).check__iterable_not_str():
            patterns = [patterns, ]

        # PATH -------------------
        if path is None:
            path = pathlib.Path.cwd()
        else:
            path = pathlib.Path(path)

        # skip_paths -------------------
        if skip_paths is None:
            skip_paths = []
        if not TypeAux(skip_paths).check__iterable_not_str():
            skip_paths = [skip_paths, ]

        for skip_path in skip_paths:
            skip_found = re.search(skip_path, str(path))
            if skip_found:
                return []

        # DIRECTORY ====================
        if path.is_dir():
            print("=" * 80)
            print(f"[PARSE FILES]")
            print(f"{path=}")
            print(f"patterns=")
            for pat in patterns:
                print(f"\t[{pat}]")
            print("-" * 80)

            result_for_dir = []
            for file in path.glob(glob_mask):
                result_for_file = cls.find_in_files(patterns=patterns, path=file, skip_paths=skip_paths)
                for item in result_for_file:
                    if item not in result_for_dir:
                        result_for_dir.append(item)

            print("-" * 80)
            print(f"[SUMMARY MATCHES]:")
            for item in  result_for_dir:
                print(f"\t{item}")
            print("=" * 80)
            return result_for_dir

        # SINGLE FILE ================
        if not path.exists():
            print(f"[not exists]{path=}")
            return []   # or raise?

        file_exc = None
        result_for_file = []
        try:
            filetext = path.read_text(encoding="utf8")
            result_for_file = cls.parse_import_modules__txt(text=filetext, patterns=patterns)
        except Exception as exc:
            file_exc = exc

        if (result_for_file or print_empty_files) or file_exc:
            print(f"{path}".ljust(80, "-"))
            if not file_exc:
                print(f"\t{result_for_file}")
            else:
                print(f"\t{file_exc}")
        return result_for_file

    # -----------------------------------------------------------------------------------------------------------------
    @classmethod
    def parse_text__import(cls, text: AnyStr) -> list[str]:
        """
        GOAL
        ----
        get imported modules from text (python sourcecode)
        get only root names (no parts like pkg.module)
        """
        return cls.parse_import_modules__txt(text, PATTERNS_IMPORT)

    @classmethod
    def parse_files__import(
            cls,
            path: Union[str, pathlib.Path] | None = None,
            print_empty: bool | None = None,
            skip_paths: Union[str, list[str]] | None = None
    ) -> list[str]:
        """
        GOAL
        ----
        get imported modules from file/filesInDirectory (python sourcecode)
        get only root names (no parts like pkg.module)

        return one cumulated list
        """
        return cls.find_in_files(patterns=PATTERNS_IMPORT, path=path, print_empty_files=print_empty, skip_paths=skip_paths)


# =====================================================================================================================
if __name__ == "__main__":
    # Packages.parse_files__import()
    Packages.find_in_files(r"Value_WithUnit", pathlib.Path.cwd().parent.parent, skip_paths=["venv", "t8", "build", ])


# =====================================================================================================================



import os

from base_aux.path2_file.m4_fileattrs import *
from base_aux.aux_text.m0_text_examples import *
from base_aux.privates.m1_privates import *
from base_aux.privates.m2_env import PvLoaderEnv


# =====================================================================================================================
def test__File():
    data_attrs = FileAttrs_Loader(text=INI_EXAMPLES.NOT_MESHED__TEXT).resolve()
    print(data_attrs)
    assert data_attrs.a0 == "00"

    data_attrs = FileAttrs_Loader(text=INI_EXAMPLES.MESHED__TEXT, keypath=("s1",)).resolve()
    print(data_attrs)
    assert data_attrs.a0 == "11"
    assert data_attrs.A0 == "11"


# =====================================================================================================================
EXAMPLE_PV_INI = """
[AUTH]
name=name1
pwd=pwd1
"""
EXAMPLE_PV_JSON = """
{"AUTH":
    {
        "name": "name1",
        "pwd": "pwd1",
    }
}
"""


class Test__PvFiles:
    VICTIM_INI = pathlib.Path.home().joinpath("test_pv.ini")
    VICTIM_JSON = pathlib.Path.home().joinpath("test_pv.json")
    @classmethod
    def setup_class(cls):
        pass
        FileAux(filepath=cls.VICTIM_INI).write__text(EXAMPLE_PV_INI)
        FileAux(filepath=cls.VICTIM_JSON).write__text(EXAMPLE_PV_JSON)

    @classmethod
    def teardown_class(cls):
        FileAux(filepath=cls.VICTIM_INI).delete_file()
        FileAux(filepath=cls.VICTIM_JSON).delete_file()

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__ini(self):
        pv = PvLoaderIni(keypath=["auth", ], filepath=self.VICTIM_INI).resolve()

        assert pv == PvLoaderIni(keypath=["AUTH", ], filepath=self.VICTIM_INI).resolve()
        assert pv != None

        assert pv.name == "name1"
        assert pv.NAME == "name1"

        # print(f"{pv_ini}")
        # print(f"{pv_ini=}")
        # print(str(pv_ini))

    def test__json(self):
        pv = PvLoaderJson(keypath=["auth", ], filepath=self.VICTIM_JSON).resolve()

        assert pv == PvLoaderJson(keypath=["AUTH", ], filepath=self.VICTIM_JSON).resolve()
        assert pv != None

        assert pv.name == "name1"
        assert pv.NAME == "name1"


# =====================================================================================================================
class Test__PvEnv:
    VALUE: str = "VALUE"

    NAME_Exists: str = "pv_Exists"
    NAME_NotExists: str = "pv_NotExists"

    @classmethod
    def setup_class(cls):
        pass
        # os.environ[cls.NAME_Exists] = cls.VALUE

    @classmethod
    def teardown_class(cls):
        pass
        # for name in [cls.NAME_Exists, cls.NAME_NotExists]:
        #     try:
        #         del os.environ[name]
        #     except:
        #         pass

    def setup_method(self, method):
        os.environ[self.NAME_Exists] = self.VALUE

    def teardown_method(self):
        for name in [self.NAME_Exists, self.NAME_NotExists]:
            try:
                del os.environ[name]
            except:
                pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__Exists(self):
        self.victim = PvLoaderEnv().resolve()

        assert self.victim[self.NAME_Exists.lower()] == self.VALUE
        assert self.victim[self.NAME_Exists.upper()] == self.VALUE

        assert self.NAME_Exists.lower() in self.victim
        assert self.NAME_Exists.upper() in self.victim

        assert getattr(self.victim, self.NAME_Exists.lower()) == self.VALUE
        assert getattr(self.victim, self.NAME_Exists.upper()) == self.VALUE

    def test__notExists(self):
        self.victim = PvLoaderEnv().resolve()

        try:
            value = self.victim[self.NAME_NotExists]
            assert False
        except Exception:
            return

    def test__len(self):
        assert len(PvLoaderEnv().resolve()) > 2
        assert len(PvLoaderEnv(patts=(r"pv.*", )).resolve()) == 1
        assert len(PvLoaderEnv(patts=(r"p1.*", )).resolve()) == 0
        assert len(PvLoaderEnv(patts=(r"p1.*", r"pv.*", )).resolve()) == 1


# =====================================================================================================================



# from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *

class Window(QWidget):
    COLUMNS = 5
    def __init__(self):
        super().__init__()

        layout_main = QHBoxLayout()
        self.setLayout(layout_main)

        font_names = QFontDatabase().families()
        col_len = len(font_names)//self.COLUMNS
        for col in range(self.COLUMNS):
            layout_col = QVBoxLayout()
            layout_main.addLayout(layout_col)

            index = 0
            while index < col_len:
                index += 1
                name = font_names.pop(0)
                wgt = QLabel(f"{name}")
                wgt.setFont(QFont(name))
                layout_col.addWidget(wgt)


if __name__ == "__main__":
    app = QApplication([])
    window = Window()
    window.show()
    app.exec_()



from base_aux.aux_eq.m1_eq_args import *
from base_aux.aux_dict.m1_dict_aux import *
from base_aux.base_types.m0_static_typing import *


# =====================================================================================================================
class Base_DiffResolve(NestInit_Args_Implicit, NestCall_Resolve):
    """
    GOAL
    ----
    get diffs from several states,
    dicts assumed like AttrDumped objects - so keys are STR.

    NOTE
    ----
    if values is ExceptionNested - apply only class!!! for both cls and inst - so next cmparing would cmp only exact classes!
    and same classes will be equal

    SPECIALLY CREATED FOR
    ---------------------
    cmp two objects by attr values
    """
    ARGS: tuple[Any, ...]
    DIFF: TYPING.DICT_ANY_TUPLE_ANY
    __diff: TYPING.DICT_ANY_TUPLE_ANY = {}

    def resolve(self) -> TYPING.DICT_ANY_TUPLE_ANY:
        """

        """
        return NotImplemented

    # -----------------------------------------------------------------------------------------------------------------
    def __bool__(self) -> bool:
        """
        GOAL
        ----
        answer the question - "Are dicts have diffs?":
            TRUE - if Diffs exists! (it looks from class name!)
            False - if NO Diffs!
        """
        return bool(self.DIFF)

    @property
    def DIFF(self) -> TYPING.DICT_ANY_TUPLE_ANY:
        """
        GOAL
        ----
        if not exists __diff value - resolve it!
        """
        if not self.__diff:
            self.resolve()
        return self.__diff

    def __str__(self) -> str:
        """
        GOAL
        ----
        print pretty result
        """
        return DictAuxInline(self.DIFF).pretty_str()


# =====================================================================================================================
class DictDiff(Base_DiffResolve):
    ARGS: tuple[TYPING.DICT_ANY_ANY, ...]

    def resolve(self) -> TYPING.DICT_ANY_TUPLE_ANY:
        keys: list[Any] = [key for DICT in self.ARGS for key in DICT]
        keys = sorted(keys)

        result = {}
        for key in keys:
            # values collect -------
            values = []
            for DICT in self.ARGS:
                if key not in DICT:
                    value = VALUE_SPECIAL.NOVALUE
                else:
                    value = DICT[key]

                if isinstance(value, Exception):    # in case of Exc() as dumped value
                    value = value.__class__

                values.append(value)

            # values check eq -------
            if not EqArgs(*values):
                result.update({key: tuple(values)})

        self.__diff = result
        return result


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from base_aux.base_types.m0_static_typing import TYPING
from base_aux.base_nest_dunders.m3_calls import *


# =====================================================================================================================
class ArgsKwargs:       # fixme: decide to separate+FINAL!!! so used only for direct KwArgs
    """
    NOTE
    ----
    1/ DONT NEST! (expect Args/Kwargs)
    2/ DONT USE!

    GOAL
    ----
    idea to keep args and kwargs in appropriate form/one object without application (constructor/cls or func).
    so we can uncovering in later.
    usage in test parametrisation.

    SPECIALLY CREATED FOR
    ---------------------
    ATC tests with using special param prefix="*"

    BEST PRACTICE
    -------------
    for item, expect in [
        (ArgsKwargs("get name"), "ATC"),
        (ArgsKwargs("test gnd", _timeout=5), "PASS"),
    ]:
        assert serialDevice.send(*item.ARGS, **item.KWARGS) == expect

    WHY NOT - 1=add direct __iter for args and smth like __dict for kwargs
    ----------------------------------------------------------------------
    and use then (*victim, **victim)
    NO - there are no __dict like dander method!
    but we can use ArgsKwargs(dict)!? - yes but it add all other methods!
        class Cls(dict):
            ARGS: tuple[Any, ...]
            KWARGS: dict[str, Any]

            def __init__(self, *args, **kwargs) -> None:
                super().__init__(**kwargs)
                self.ARGS = args
                self.KWARGS = kwargs

            def __iter__(self) -> Iterator[Any]:
                yield from self.ARGS

    so as result the best decision is (*item.ARGS, **item.KWARGS)
    and we could use this class as simple base for Lambda for example!
    """
    ARGS: TYPING.ARGS_FINAL = ()
    KWARGS: TYPING.KWARGS_FINAL = {}

    def __init__(self, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs

    def __bool__(self) -> bool:
        if self.ARGS or self.KWARGS:
            return True
        else:
            return False


# ---------------------------------------------------------------------------------------------------------------------
# @final        # DONT USE final! need
class Args(ArgsKwargs, NestCall_Resolve):
    """
    just a derivative to clearly show only Args is important
    """
    def __bool__(self) -> bool:
        if self.ARGS:
            return True
        else:
            return False

    def __iter__(self):
        yield from self.ARGS

    def __contains__(self, item):
        return item in self.ARGS

    def resolve(self):
        return self.ARGS


# ---------------------------------------------------------------------------------------------------------------------
@final
class Kwargs(dict):
    """
    NOTE
    ----
    for combine clear kwargs use direct DICT(key1=1, ...) instead!

    WHY NOT - 1: DIRECT DICT(key1=1, ...)
    -------------------------------------
    you are write! use it!
    """
    # def __bool__(self) -> bool:
    #     if self.KWARGS:
    #         return True
    #     else:
    #         return False
    #
    # def __iter__(self):
    #     yield from self.KWARGS
    #
    # def __call__(self):
    #     return self.KWARGS


# =====================================================================================================================



from typing import *
import pytest
import time
import threading
import abc

from base_aux.base_singletons.m1_singleton import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="Victim",
    argvalues=[
        Singleton_CallMeta,
        Singleton_New
    ]
)
def test__no_args(Victim):
    class Victim1(Victim):
        attr = 1

    class Victim2(Victim):
        attr = 2

    Victim._drop_all()
    assert Victim._SINGLETONS == []

    assert Victim1().attr == 1
    Victim1().attr = 11
    assert Victim1().attr == 11
    assert Victim._SINGLETONS == [Victim1(), ]
    assert Victim1._SINGLETONS == [Victim1(), ]
    try:
        assert Victim1()._SINGLETONS == [Victim1(), ]
    except:
        assert Victim == Singleton_CallMeta
    else:
        assert Victim == Singleton_New

    assert Victim2().attr == 2
    Victim2().attr = 22
    assert Victim2().attr == 22
    assert Victim._SINGLETONS == [Victim1(), Victim2()]
    assert Victim2._SINGLETONS == [Victim1(), Victim2()]
    try:
        assert Victim2()._SINGLETONS == [Victim1(), Victim2()]
    except:
        assert Victim == Singleton_CallMeta
    else:
        assert Victim == Singleton_New

    assert Victim1().attr == 11


def test__META_with_args():
    Victim = Singleton_CallMeta
    Victim._drop_all()
    class Victim1(Victim):
        def __init__(self, attr):
            self.attr = attr

    assert Victim._SINGLETONS == []
    instance = Victim1(1)

    assert instance.attr == 1
    assert Victim1(111).attr == 1
    assert Victim._SINGLETONS == [instance, ]

    Victim1(111).attr = 11
    assert Victim1(1).attr == 11
    assert Victim._SINGLETONS == [instance, ]

    assert Victim1(1111).attr == 11
    assert Victim._SINGLETONS == [instance, ]


def test__NoMETA_with_args():
    Victim = Singleton_New
    Victim._drop_all()
    class Victim1(Victim):
        def __init__(self, attr):
            self.attr = attr

    assert Victim._SINGLETONS == []
    instance = Victim1(1)

    assert instance.attr == 1
    assert Victim1(111).attr == 111
    assert Victim._SINGLETONS == [instance, ]

    assert Victim1(1).attr == 1
    assert Victim._SINGLETONS == [instance, ]


def test__META_nesting_else_one_meta():
    # cant use else one metaclass in nesting!
    try:
        class Victim1(Singleton_CallMeta, abc.ABC):
            attr = 1
    except TypeError:
        msg = """
        TypeError: metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all its bases

        """
        pass
    else:
        assert False

    class Victim2(Singleton_New, abc.ABC):
        attr = 2
    assert Victim2().attr == 2
    Victim2().attr = 22
    assert Victim2().attr == 22


@pytest.mark.parametrize(argnames="Victim", argvalues=[Singleton_CallMeta, Singleton_New])
def test__threading_spawn(Victim):
    class Victim1(Victim):
        def __init__(self):
            time.sleep(1)

    threads = [
        threading.Thread(target=Victim1),
        threading.Thread(target=Victim1),
        threading.Thread(target=Victim1),
        threading.Thread(target=Victim1),
        threading.Thread(target=Victim1),
    ]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()
        assert thread.is_alive() is False


@pytest.mark.parametrize(argnames="Victim", argvalues=[Singleton_CallMeta, Singleton_New])
def test__several_levels_at_ones__low(Victim):
    class VictimBase2(Victim):
        attr = 0
    class Victim1(VictimBase2):
        attr = 1
    class Victim2(VictimBase2):
        attr = 2

    assert VictimBase2().attr == 0
    try:
        assert Victim1().attr == 1
    except Exc__NestingLevels:
        pass
    else:
        assert False


@pytest.mark.parametrize(argnames="Victim", argvalues=[Singleton_CallMeta, Singleton_New])
def test__several_levels_at_ones__up(Victim):
    class VictimBase2(Victim):
        attr = 0

    class Victim1(VictimBase2):
        attr = 1

    class Victim2(VictimBase2):
        attr = 2

    assert Victim1().attr == 1
    assert Victim2().attr == 2
    try:
        assert VictimBase2().attr == 0
    except Exc__NestingLevels:
        pass
    else:
        assert False


# =====================================================================================================================



import pytest

from base_aux.aux_attr.m3_ga0_echo import *


# =====================================================================================================================
def test__GetattrEcho():
    assert NestGaCls_Echo.hello == "hello"
    assert NestGaCls_Echo.Hello == "Hello"
    assert NestGaCls_Echo.ПРИВЕТ == "ПРИВЕТ"

    assert NestGaCls_Echo.hello_world == "hello_world"


def test__GetattrEchoSpace():
    assert NestGaCls_EchoSpace.hello == "hello"
    assert NestGaCls_EchoSpace.Hello == "Hello"
    assert NestGaCls_EchoSpace.ПРИВЕТ == "ПРИВЕТ"

    assert NestGaCls_EchoSpace.hello_world == "hello world"
    assert NestGaCls_EchoSpace.hello__world == "hello  world"


# =====================================================================================================================



import pytest
import asyncio

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_types.m0_static_typing import *
from base_aux.aux_datetime.m1_sleep import *


# =====================================================================================================================
@pytest.mark.asyncio
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        # TRIVIAL ---------------------------------
        (None, Exception),
        (1, Exception),
        (Base_SleepAw(0.1), Exception),
        (SleepAwNone(0.1), None),
        (SleepAwTrue(0.1), True),
        (SleepAwFalse(0.1), False),
        (SleepAwExc(0.1), Exception),
        (SleepAwRaise(0.1), Exception),
    ]
)
async def test__1(
        source: TYPING.AIO_CORO | Any,
        _EXPECTED: bool | type[Exception] | Any,
):
    try:
        result = await source
    except BaseException as exc:
        result = exc

    Lambda(result).check_expected__assert(_EXPECTED)


# =====================================================================================================================



import asyncio
from typing import *
import pathlib
import collections

from base_aux.base_values.m2_value_special import VALUE_SPECIAL
from base_aux.base_types.m0_static_types import TYPES
# from base_aux.base_lambdas.m1_lambda import Lambda    # DONT IMPORT! circular inport!


# =====================================================================================================================
@final
class TYPING:
    """
    GOAL
    ----
    collect all typing USER variants
    """
    ELEMENTARY = Union[*TYPES.ELEMENTARY]

    # DIGIT -----------------------------------------------------------------------------------------------------------
    DIGIT_FLOAT_INT = Union[float, int]
    DIGIT_FLOAT_INT_NONE = Union[float, int, None]

    # ARGS/KWARGS -----------------------------------------------------------------------------------------------------
    ARGS_FINAL = tuple[Any, ...]
    ARGS_DRAFT = Union[Any, ARGS_FINAL, 'ArgsKwargs']           # you can use direct single value

    KWARGS_FINAL = dict[str, Any]
    KWARGS_DRAFT = Union[None, KWARGS_FINAL, 'ArgsKwargs']  # if passed NONE - no data!

    # PATH ------------------------------------------------------------------------------------------------------------
    PATH_FINAL = pathlib.Path
    PATH_DRAFT = Union[str, PATH_FINAL]

    STR_FINAL = str
    STR_DRAFT = Union[STR_FINAL, Any]

    ATTR_FINAL = str
    ATTR_DRAFT = str | int | Any

    # DICT ------------------------------------------------------------------------------------------------------------
    DICT_ANY_NONE = dict[Any, None]             # just to show - dict with None values after clearing!
    DICT_ANY_ANY = dict[Any, Any]               # just to show - dict could be any! on keys/values
    DICT_STR_ANY = dict[str, Any]               # just to show - dict could be any! on values! not just an elementary1
    DICT_STR_ELEM = DICT_JSON_ANY = dict[str, ELEMENTARY]       # just to show - parsed by json - dict
    DICT_STR_STR = DICT_INI = dict[str, str]               # just to show - parsed by ini!
    JSON_ANY = ELEMENTARY | DICT_STR_ELEM  # just to show - parsed by json - any object

    DICT_ANY_TUPLE_ANY = dict[Any, tuple[Any, ...]]     # as DICT_DIFF_DICTS
    DICT_STR_TUPLE_ANY = dict[str, tuple[Any, ...]]     # as DICT_DIFF_ATTRS

    # -----------------------------------------------------------------------------------------------------------------
    ITERABLE_ORDERED = Union[dict, list, tuple, Iterable]     # "SET" - DONT USE!
    ITERPATH_KEY = Union[Any, int]   # Any is for dict
    ITERPATH = tuple[ITERPATH_KEY, ...]

    # -----------------------------------------------------------------------------------------------------------------
    BOOL_DRAFT = Union[
        None,
        Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
        bool,                               # as main idea! as already final generic
        Exception,
        Callable[..., bool | Any | NoReturn | Exception],   # as main idea! to get final generic
        VALUE_SPECIAL.NOVALUE
    ]

    # RESULT ----------------------------------------------------------------------------------------------------------
    RESULT__NONE = None
    RESULT__BOOL = bool
    RESULT__EXC = Exception     # | type[Exception] - DONT USE typeExc! from any process Exception comes as instance!

    RESULT__ANY_EXC = Any | Exception
    RESULT__ANY_RAISE = Any | NoReturn

    RESULT__BOOL_NONE = bool | None
    RESULT__BOOL_RAISE = bool | NoReturn
    RESULT__BOOL_EXC = bool | Exception
    RESULT__RAISE_NONE = NoReturn | None

    RESULT__BOOL_RAISE_NONE = bool | NoReturn | None

    # -----------------------------------------------------------------------------------------------------------------
    EXPECTED = bool | type[Any | Exception] | Any

    # CALLABLE --------------------------------------------------------------------------------------------------------
    CALLABLE_DRAFT = Union[Any, type[Any], Callable[..., Any | NoReturn]]
    # CALLABLE_FINAL    # dont need final! Final - are all others!

    CALLABLE__NONE = Callable[..., None]
    CALLABLE__BOOL = Callable[..., bool]
    CALLABLE__EXC = Callable[..., Exception]

    CALLABLE__BOOL_NONE = Callable[..., bool | None]
    CALLABLE__BOOL_RAISE = Callable[..., bool | NoReturn]
    CALLABLE__BOOL_EXC = Callable[..., bool | Exception]
    CALLABLE__RAISE_NONE = Callable[..., NoReturn | None]  # not expecting any bool! intended/inportant only raising as inappropriate position!

    CALLABLE__BOOL_RAISE_NONE = Callable[..., bool | NoReturn | None]

    # -----------------------------------------------------------------------------------------------------------------
    VALID_VALIDATOR = Union[
        Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
        # Type,   # Class as validator like Exception????? fixme
        type[Exception],  # direct comparison
        Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
    ]

    # -----------------------------------------------------------------------------------------------------------------
    AIO_FUNC = Callable[..., Awaitable[Any]]                    # func defined with ASYNC key which result as AWAITABLE
    AIO_FUNC__ANY = Callable[..., Awaitable[Any]]                    # func defined with ASYNC key which result as AWAITABLE
    AIO_FUNC__BOOL = Callable[..., Awaitable[bool]]                    # func defined with ASYNC key which result as AWAITABLE

    AIO_CORO = collections.abc.Coroutine[Any, Any, bool | Any]  # as new version for typing
    AIO_CORO__ANY = collections.abc.Coroutine[Any, Any, Any]
    AIO_CORO__BOOL = collections.abc.Coroutine[Any, Any, bool]

    AIO_AW = Awaitable[Any]     # object
    AIO_AW__ANY = Awaitable[Any]
    AIO_AW__BOOL = Awaitable[bool]

    AIO_TASK = asyncio.Task     # started Task in the loop
    """
    import asyncio
    import types
    
    async def example():
        pass
    
    print(type(example))           # <class 'function'>
    print(type(example()))         # <class 'coroutine'>
    print(isinstance(example, types.CoroutineType))  # False
    print(isinstance(example(), types.CoroutineType)) # True
    """


# =====================================================================================================================



import configparser

from base_aux.base_types.m0_static_types import *
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class ConfigParserMod(configparser.ConfigParser):
    """
    GOAL
    ----
    just add some dict-meths into original class
    """
    def to_dict__direct(self) -> TYPING.DICT_STR_STR | None:
        """
        NOTE
        ----
        it is just an ability!
        use final MERGED and KEEP SOURCE in CORRECT struct! NoDEFAULTS or know what you do!!
        """
        # use double places!
        result = dict(self._defaults)
        if result:
            result.update({"DEFAULT": dict(self._defaults)})

        # sections
        result.update(dict(self._sections))
        if not result:
            return
        return result

    def to_dict__merged(self) -> TYPING.DICT_STR_STR:
        result = self.to_dict__direct()
        for section in self.sections():
            for def_name, def_val in self._defaults.items():
                if def_name not in result[section]:
                    result[section][def_name] = def_val
        return result

    def to_dict(self) -> TYPING.DICT_STR_STR:
        return self.to_dict__merged()

    # -----------------------------------------------------------------------------------------------------------------
    def read_string(self, data: str, *args, **kwargs) -> None | NoReturn:
        """
        GOAL-mod
        ----
        just fix default section
        """
        try:
            super().read_string(string=data, *args, **kwargs)
        except:
            data = f"[DEFAULT]\n{data}"
            super().read_string(string=data, *args, **kwargs)


# =====================================================================================================================
# def _explore():
#     config = ConfigParserMod()
#     config.set("DEFAULT", "name0", "000")
#     config.add_section("SEC1")
#     config.set("SEC1", "name1", "111")
#     config.set("SEC1", "name2", "222")
#     config.set("SEC1", "name3", "333")
#     print(config.to_dict__direct())
#     print(config.to_dict__merged())


# =====================================================================================================================
# if __name__ == '__main__':
#     _explore()


# =====================================================================================================================



from typing import *
from bs4 import BeautifulSoup
from base_aux.alerts.m2_select import *
from base_aux.alerts.m1_alert0_base import *
from base_aux.base_types.m2_info import *
from base_aux.chains.m1_chains import *


# =====================================================================================================================
# TODO: make tag access by attr name with calling and selection by kwargs
#   tag.a(**kwargs).header(**kwargs).b().a()


# =====================================================================================================================
class HtmlTagParser(NestCall_Resolve, NestInit_AnnotsAttr_ByKwargs):
    """
    GOAL
    ----
    get value(body/load) for exact html tag from source.
    if several tags found - return exact index!

    KWARGS - all params directly will be passed into function Tag.find_all!

    EXAMPLES
    --------
    HtmlTagAddress(name="table", attrs={"class": "donor-svetofor-restyle"}),
    """
    SOURCE: str = None
    INDEX: int = 0

    # BS4.find_all params ---------------------------------------------------------------------------------------------
    NAME: str = None
    ATTRS: dict[str, str] = dict()
    STRING: Optional[str] = None
    RECURSIVE: bool = None

    def __init__(self, source: str | BeautifulSoup = NoValue, **kwargs) -> None:
        if source not in [NoValue, None]:
            self.SOURCE = str(source)

        super().__init__(**kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self, source: str | BeautifulSoup = NoValue) -> None | str:
        if source not in [NoValue, None]:
            self.SOURCE = str(source)

        return self.get_body()

    # -----------------------------------------------------------------------------------------------------------------
    def get_body(self) -> None | str:
        if self.SOURCE:
            try:
                bs_tag: BeautifulSoup = BeautifulSoup(markup=self.SOURCE, features='html.parser')
            except Exception as exc:
                msg = f"can't parse {self.SOURCE=}\n{exc!r}"
                Warn(msg)
                return
        else:
            msg = f"empty {self.SOURCE=}"
            Warn(msg)
            return

        # collect params ----------------------------------------------------------------------------------------------
        params = dict()
        if self.NAME is not None:
            params |= dict(name=self.NAME)
        if self.ATTRS:
            params |= dict(attrs=self.ATTRS)
        if self.STRING is not None:
            params |= dict(string=self.STRING)
        if self.RECURSIVE is not None:
            params |= dict(recursive=self.RECURSIVE)

        params |= dict(limit=self.INDEX + 1)

        # find --------------------------------------------------------------------------------------------------------
        try:
            tags = bs_tag.find_all(**params)
            bs_tag = tags[self.INDEX]
        except Exception as exc:
            msg = f"URL WAS CHANGED? can't find {self=}\n{exc!r}"
            Warn(msg)
            return

        return bs_tag.decode_contents()     # get exact internal boby(load) of tag without self-bracket-markup


# =====================================================================================================================
class ChainResolve_HtmlTagParser(ChainResolve):
    ON_RAISE: EnumAdj_ReturnOnRaise = EnumAdj_ReturnOnRaise.NONE


# =====================================================================================================================
def explore():
    load0 = 'hello<a href="http://example.com/">\nlink <i>example.com</i>\n</a>'

    # soup = BeautifulSoup(markup, 'html.parser')
    # print(soup.a)
    # print()
    # print()
    # print()
    # ObjectInfo(soup.a).print()
    #
    # print()
    # print()
    # print()

    # for name in dir(soup.a):
    #     print(name)

    # load = HtmlTagParser(load0, name="a").resolve()
    # print(f"{load=}")
    # load = HtmlTagParser(load0, name="i").resolve()
    # print(f"{load=}")

    print(f"{ChainResolve_HtmlTagParser(HtmlTagParser(name="a"), source=load0).resolve()=}")
    print(f"{ChainResolve_HtmlTagParser(HtmlTagParser(name="a")).resolve(source=load0)=}")
    print(f"{ChainResolve_HtmlTagParser(HtmlTagParser(name="a"), HtmlTagParser(name="i")).resolve(source=load0)=}")
    print(f"{ChainResolve_HtmlTagParser(HtmlTagParser(name="a"), HtmlTagParser(name="i"), source=load0).resolve()=}")


# =====================================================================================================================
if __name__ == "__main__":
    explore()
    # run_over_api()


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m5_value_valid1_eq import *
from base_aux.aux_eq.m4_eq_valid_chain import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, eq, _EXPECTED",
    argvalues=[
        (LAMBDA_RAISE, NoValue, None),
        (1, NoValue, None),

        (1, EqValid_GE(1), None),
        (1, EqValid_GE(2), Exception),

        (1, EqValid_EQ(1), None),
        (1, EqValid_EQ(2), Exception),
    ]
)
def test__1_init(source, eq, _EXPECTED):
    if _EXPECTED is None:
        _EXPECTED = source
        # here it will be compared with source or Exception!

    func_link = ValueEqValid
    Lambda(func_link, source, eq).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, eq, new, _EXPECTED",
    argvalues=[
        (LAMBDA_RAISE, NoValue, 10, True),
        (1, NoValue, 10, True),

        (1, EqValid_GE(1), 10, True),
        (1, EqValid_GE(1), 0, Exception),

        (1, EqValid_EQ(1), 10, Exception),
        (1, EqValid_EQ(1, 10), 10, True),
    ]
)
def test__2_reset(source, eq, new, _EXPECTED):
    func_link = ValueEqValid(source, eq).value_update
    Lambda(func_link, new).check_expected__assert(_EXPECTED)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, eq, other, _EXPECTED",
    argvalues=[
        (1, NoValue, 1, True),
        (1, NoValue, 10, False),

        (1, EqValid_GE(1), 1, True),
        (1, EqValid_GE(1), 10, False),
        (1, EqValid_GE(1), 0, False),

        (1, EqValid_EQ(1), 10, False),
        (1, EqValid_EQ(1, 10), 10, False),
    ]
)
def test__3_eq(source, eq, other, _EXPECTED):
    func_link = ValueEqValid(source, eq) == other
    Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_text.m4_ini import ConfigParserMod
from base_aux.aux_text.m0_text_examples import *


# =====================================================================================================================
class Test__Ini:
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, (None, None)),
            ("", (None, None)),
            ("1", Exception),
            (INI_EXAMPLES.INT_KEY__TEXT, (INI_EXAMPLES.INT_KEY__DICT_DIRECT, INI_EXAMPLES.INT_KEY__DICT_DIRECT)),
            (INI_EXAMPLES.NOT_MESHED__TEXT, (INI_EXAMPLES.NOT_MESHED__DICT_DIRECT, INI_EXAMPLES.NOT_MESHED__DICT_MERGED)),
            (INI_EXAMPLES.MESHED__TEXT, (INI_EXAMPLES.MESHED__DICT_DIRECT, INI_EXAMPLES.MESHED__DICT_MERGED)),
        ]
    )
    def test__to_dict(self, source, _EXPECTED):
        victim = ConfigParserMod()

        try:
            victim.read_string(source)
        except Exception as exc:
            Lambda(exc).check_expected__assert(_EXPECTED)
            return

        Lambda(victim.to_dict__direct).check_expected__assert(_EXPECTED[0])
        Lambda(victim.to_dict__merged).check_expected__assert(_EXPECTED[1])


def _explore():
    victim = ConfigParserMod()
    victim.set("DEFAULT", "n0", "000")
    victim.add_section("SEC1")
    victim.set("SEC1", "n1", "111")
    result = victim.to_dict__direct()
    print(victim.to_dict__merged())


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class Base_AttrDumped:
    """
    SPECIALLY CREATED FOR
    ---------------------
    make obj with independent __annotations__
    """
    pass

    # TODO: how to add DumpStrPretty here without mess attr names???


@final
class AttrDumped(Base_AttrDumped):
    """
    GOAL
    ----
    just use static bare class for dumping any set of attrs!

    WHY NOT - AttrsKit
    ------------------
    cause sometimes it makes circular recursion exc!
    """


# =====================================================================================================================
if __name__ == "__main__":
    # OK
    class Cls:
        pass

    print(Cls.__annotations__)      # class first - ok! like touch!
    print(Cls().__annotations__)
    print()

    # FAIL
    class Cls:
        pass

    # print(Cls().__annotations__)    # inst first - exc
    print(Cls.__annotations__)
    print()

    class Cls:
        pass

    victim = Cls()
    try:
        print(victim.__annotations__)
    except:
        pass
    else:
        assert False
    victim.__class__.__annotations__
    print(victim.__annotations__)

    victim.__class__.__annotations__.update(attr=1)
    assert victim.__annotations__ != victim.__class__.__annotations__


# =====================================================================================================================



from typing import *

from .tc0__base import *


# =====================================================================================================================
class TestCase(Base_TcAtcPtb):
    ATC_VOUT: int | None = 220
    PTB_SET_EXTON: bool = False
    PTB_SET_HVON: bool = True
    PTB_SET_PSON: bool = True

    ASYNC = True
    _DESCRIPTION = "тест нагрузочный\nшины главного питания (MAIN)"

    def _gen_chains(self, load: int) -> list[Valid]:
        return [
            ValidFailStop(
                value_link=self.DEV_COLUMN.DUT.SET,
                kwargs__value={"LOAD": load},
                validate_link="OK",
                name="SET",
            ),
            ValidSleep(1),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="VINOK",
                validate_link=1,
            ),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="PWOK",
                validate_link=1,
            ),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="V12S",
                validate_link=EqValid_LGTE_NumParsedSingle(ge=11, le=13).resolve,
                name="GET+valid diapason",
            ),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="V12M",
                validate_link=EqValid_LGTE_NumParsedSingle(ge=11, le=13).resolve,
                name="GET+valid diapason",
            ),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="IOUT",
                validate_link=EqValid_LGTE_NumParsedSingle(ge=load / 12 / 2).resolve,
                name="GET+valid diapason",
            ),
            ValidNoCum(
                value_link=self.DEV_COLUMN.DUT.GET,
                args__value="IIN",
                validate_link=EqValid_LGTE_NumParsedSingle(ge=load / 12 / 2).resolve,
                name="GET+valid diapason",
            ),
        ]

    # -----------------------------------------------------------------------------------------------------------------
    def run__wrapped(self) -> TYPING__RESULT_W_EXC:
        result_chain = ValidChains(
            chains=[
                *self._gen_chains(0),
                *self._gen_chains(100),
                *self._gen_chains(300),
                *self._gen_chains(400),
                *self._gen_chains(800),

                *self._gen_chains(0),
                *self._gen_chains(800),
            ]
        )
        return result_chain


# =====================================================================================================================



import time
import pytest


# =====================================================================================================================
# KEEP FILES IN ROOT! OR IMPORT PRJ_MODULE WOULD FROM SYSTEM! NOT THIS SOURCE!!!
from base_aux.loggers.m2_logger import *


# =====================================================================================================================
class Test__1:
    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass
        # self.VICTIM = type("VICTIM", (Logger,), {})

    def teardown_method(self, method):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__maxRecursion(self):
        class Victim(Logger):
            LOG_ENABLE = True
            pass

        victim = Victim()
        for i in range(3):
            victim.LOGGER.debug(f"hello-{i}")

        assert True

    def test__disable(self):
        class VictimDisabled(Logger):
            LOG_ENABLE = False

        victim = VictimDisabled()
        victim.LOGGER.debug(f"VictimDisabled")

        time.sleep(0.2)

        assert not victim.LOG_FILEPATH.exists()

    def test__several_same_creations(self):
        # in case of starting several times like in ClientRequestsItem! see there!
        class Victim(Logger):
            LOG_ENABLE = True

        victim1 = Victim()
        assert len(victim1.LOGGER.handlers) == 1
        victim2 = Victim()
        assert len(victim2.LOGGER.handlers) == 1
        victim3 = Victim()
        assert len(victim3.LOGGER.handlers) == 1


# =====================================================================================================================



from base_aux.aux_text.m5_re1_rexp import *
from base_aux.aux_iter.m1_iter_aux import *
from base_aux.base_values.m3_exceptions import *


# =====================================================================================================================
TYPING__OTHER_DRAFT = str | Any
TYPING__RE_RESULT__ONE = str | tuple[str, ...]
TYPING__RE_RESULT__ALL = TYPING__RE_RESULT__ONE | list[TYPING__RE_RESULT__ONE]


# =====================================================================================================================
class Base_ReAttempts:
    """
    GOAL
    ----
    apply same methods as in RE module, but
    work with attempts

    NOTE
    ----
    ATTEMPTS_USAGE
    if FIRST - return result for first match function in attempt order
    if ALL - return list of results for all matched attempts
    """
    ATTEMPTS: TYPING__REXPS_FINAL
    FLAGS_DEF: int = None
    ATTEMPTS_USAGE: EnumAdj_AttemptsUsage = EnumAdj_AttemptsUsage.ALL

    def __init__(self, *attempts: TYPING__REXP_DRAFT, flags_def: int = None, attempts_usage: EnumAdj_AttemptsUsage = None) -> None:
        if flags_def is not None:
            self.FLAGS_DEF = flags_def

        if attempts_usage is not None:
            self.ATTEMPTS_USAGE = EnumAdj_AttemptsUsage(attempts_usage)

        result = []
        for attempt in attempts:
            if isinstance(attempt, RExp):
                result.append(attempt)
            elif isinstance(attempt, str):
                result.append(RExp(attempt))
            else:
                raise Exc__Incompatible_Data(f"{attempt=}")

        self.ATTEMPTS = result

    # -----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _result__get_from_match(match: re.Match) -> TYPING__RE_RESULT__ONE:
        """
        NOTE
        ----
        this is one of the the main idea for whole this class!

        GOAL
        ----
        get result from match object
        1. if no groups - return matching string
        2. if one group - return exact the group value
        3. if several groups - return tuple of groups
        """
        if not isinstance(match, re.Match):
            raise Exc__WrongUsage(f"{match=}")

        groups = match.groups()
        if groups:
            if len(groups) == 1:
                return groups[0]
            else:
                return groups
        else:
            return match.group()

    # -----------------------------------------------------------------------------------------------------------------
    pass

    # return None only for FIRST! if ALL - retrun always LIST!

    def match(self, other: TYPING__OTHER_DRAFT) -> TYPING__RE_RESULT__ALL | None:
        other = str(other)
        result = []
        for rexp in self.ATTEMPTS:
            flags = IterAux([rexp.FLAGS, self.FLAGS_DEF, 0]).get_first_is_not_none()

            match = re.match(rexp.PAT, other, flags)
            if match:
                result_i = self._result__get_from_match(match)
                if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
                    return result_i
                else:
                    result.append(result_i)

        # finish
        if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
            return None
        else:
            return result

    def fullmatch(self, other: TYPING__OTHER_DRAFT) -> TYPING__RE_RESULT__ALL | None:
        other = str(other)
        result = []
        for rexp in self.ATTEMPTS:
            flags = IterAux([rexp.FLAGS, self.FLAGS_DEF, 0]).get_first_is_not_none()

            match = re.fullmatch(rexp.PAT, other, flags)
            if match:
                result_i = self._result__get_from_match(match)
                if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
                    return result_i
                else:
                    result.append(result_i)
        # finish
        if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
            return None
        else:
            return result

    def search(self, other: TYPING__OTHER_DRAFT) -> TYPING__RE_RESULT__ALL | None:
        other = str(other)
        result = []
        for rexp in self.ATTEMPTS:
            flags = IterAux([rexp.FLAGS, self.FLAGS_DEF, 0]).get_first_is_not_none()

            match = re.search(rexp.PAT, other, flags)
            if match:
                result_i = self._result__get_from_match(match)
                if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
                    return result_i
                else:
                    result.append(result_i)
        # finish
        if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
            return None
        else:
            return result

    # -----------------------------------------------------------------------------------------------------------------
    def findall(self, other: TYPING__OTHER_DRAFT) -> list[TYPING__RE_RESULT__ONE]:
        other = str(other)
        result = []
        for rexp in self.ATTEMPTS:
            flags = IterAux([rexp.FLAGS, self.FLAGS_DEF, 0]).get_first_is_not_none()

            result_i = re.findall(rexp.PAT, other, flags)
            if result_i:
                if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
                    return result_i
                else:
                    result.extend(result_i)

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def sub(self, other: TYPING__OTHER_DRAFT, new: str = None) -> str:
        other = str(other)
        result = other
        for rexp in self.ATTEMPTS:
            flags = IterAux([rexp.FLAGS, self.FLAGS_DEF, 0]).get_first_is_not_none()
            new = IterAux([rexp.SUB, new, ""]).get_first_is_not_none()
            count = IterAux([rexp.SCOUNT, 0]).get_first_is_not_none()

            result = re.sub(rexp.PAT, new, other, count, flags)
            if result != other:
                other = result
                if self.ATTEMPTS_USAGE == EnumAdj_AttemptsUsage.FIRST:
                    break

        return result

    def delete(self, other: TYPING__OTHER_DRAFT) -> str:
        return self.sub(other)


# =====================================================================================================================
@final
class ReAttemptsFirst(Base_ReAttempts):
    ATTEMPTS_USAGE = EnumAdj_AttemptsUsage.FIRST


@final
class ReAttemptsAll(Base_ReAttempts):
    ATTEMPTS_USAGE = EnumAdj_AttemptsUsage.ALL


# =====================================================================================================================



import time

from base_aux.testplans.tc import *
from base_aux.valid.m2_valid_derivatives import *
from base_aux.valid.m3_valid_chains import *
from .tc0_groups import *


# =====================================================================================================================
class TestCase(Base_TestCase):
    ASYNC = True
    DESCRIPTION = "direct1"

    @classmethod
    @property
    def _EQ_CLS__VALUE(cls) -> Enum_TcGroup:
        return Enum_TcGroup.G1

    def run__wrapped(self):
        time.sleep(0.1)
        self.details_update({"detail_value": self.DEV_COLUMN.DUT.VALUE})
        result_chain = ValidChains(
            [
                Valid(value_link=True, name="TRUE"),
            ],
        )
        return result_chain


# =====================================================================================================================



from base_aux.loggers.m1_print import Warn


# =====================================================================================================================
# USE COMMON/GENERAL TYPES

_std = [
    # base ----------------
    AssertionError,

    # FILE/PATH
    NotADirectoryError,
    IsADirectoryError,

    # USER ----------------
    UserWarning,
    Warning,
    DeprecationWarning,
    PendingDeprecationWarning,

    InterruptedError,

    NotImplemented,
    NotImplementedError,

    # VALUE ---------------
    TypeError,      # type
    ValueError,     # value

    # ACCESS ------
    PermissionError,

    # COLLECTION
    GeneratorExit,
    StopIteration,
    StopAsyncIteration,

    # arithm/logic
    ZeroDivisionError,
    ArithmeticError,
    FloatingPointError,
    OverflowError,

    RecursionError,
    BrokenPipeError,

    # OS/OTHER
    SystemExit,
    # WindowsError,     # NOTE: NOT EXISTS IN LINUX!!! dont use in any variation!!!
    IOError,
    OSError,
    EnvironmentError,
    SystemError,
    ChildProcessError,
    MemoryError,
    KeyboardInterrupt,

    BufferError,
    LookupError,

    UnboundLocalError,

    # PROCESS
    RuntimeWarning,
    ResourceWarning,
    ReferenceError,
    ProcessLookupError,
    RuntimeError,
    FutureWarning,
    ExceptionGroup,
    BlockingIOError,

    # REAL VALUE = NOT AN EXCEPTION!!!
    NotImplemented,      # NotImplemented = None # (!) real value is 'NotImplemented'
]


# =====================================================================================================================
class Base_Exc(
    Warn,

    Exception,
    # BaseException,
    # BaseExceptionGroup,
):
    """
    GOAL
    ----
    1/ with raise - just a solution to collect all dunder methods intended for Exceptions in one place
        - get correct bool() if get Exc as value
    2/ without raising - use like logger (Warn)

    SPECIALLY CREATED FOR
    ---------------------
    classes.VALID if
    """
    PREFIX: str = "[EXC]"


# =====================================================================================================================
class Exc__EncodeDecode(
    Base_Exc,

    # BytesWarning,
    # EncodingWarning,
    # UnicodeWarning,
    # UnicodeDecodeError,
    # UnicodeEncodeError,
    # UnicodeTranslateError,
):
    """
    GOAL
    ----
    collect all EncodeDecode Errors
    """
    pass


class Exc__Connection(
    Base_Exc,

    # ConnectionError,
    # ConnectionAbortedError,
    # ConnectionResetError,
    # ConnectionRefusedError,
    # TimeoutError,
):
    pass


class Exc__Imports(
    Base_Exc,

    # ImportError,
    # ImportWarning,
    # ModuleNotFoundError,
):
    pass


class Exc__SyntaxFormat(
    Base_Exc,

    # SyntaxWarning,
    # SyntaxError,
    # IndentationError,
    #
    # EOFError,
    # TabError,
):
    pass


class Exc__Addressing(
    Base_Exc,

    # NameError,
    # AttributeError,
    # KeyError,
    # IndexError,
):
    pass


class Exc__NotExistsNotFoundNotCreated(
    Base_Exc,

    # FileExistsError,    # ExistsAlready
    # FileNotFoundError,  # NotExists
):
    """
    GOAL
    ----
    any exception intended Exists/NotExists any object
    dont mess with ADDRESSING!
    """
    pass


# =====================================================================================================================
class Exc__Incompatible(Base_Exc):
    """
    GOAL
    ----
    any incompatibility
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
class Exc__Incompatible_Data(Exc__Incompatible):
    pass


class Exc__Incompatible_Struct(Exc__Incompatible):
    """
    GOAL
    ----
    data has incorrect structure/schema

    SPECIALLY CREATED FOR
    ---------------------
    aio_testplan.TableObj._validate_schema
    """
    pass


# =====================================================================================================================
class Exc__WrongUsage(Base_Exc):
    """
    GOAL
    ----
    somebody perform incorrect usage!
    """


# ---------------------------------------------------------------------------------------------------------------------
class Exc__WrongUsage_Programmer(Exc__WrongUsage):
    """
    GOAL
    ----
    wrong programmer behaviour (smth about architecture)
    """
    pass


class Exc__WrongUsage_YouForgotSmth(Exc__WrongUsage_Programmer):
    """
    GOAL
    ----
    just a shallow error when you forget smth


    SPECIALLY CREATED FOR
    ---------------------
    ReleaseHistory - cause it is not Programmer
    """
    pass


# =====================================================================================================================
class Exc__Expected(Base_Exc):
    """
    GOAL
    ----
    Any requirement/exact cmp/eq
    """


class Exc__Overlayed(Base_Exc):
    """
    GOAL
    ----
    ENY OVERLAY ITEMS/ADDRESSES
    index
    """
    pass


class Exc__NotReady(Base_Exc):
    """
    GOAL
    ----
    any not ready state/object
    connection/...
    """
    pass


# =====================================================================================================================
class Exc__GetattrPrefix(Base_Exc):
    pass


class Exc__GetattrPrefix_RaiseIf(Exc__GetattrPrefix):
    pass


class Exc__StartOuterNONE_UsedInStackByRecreation(Base_Exc):
    """
    in stack it will be recreate automatically! so dont use in pure single BreederStrSeries!
    """
    pass


# =====================================================================================================================
class Exc__NestingLevels(Base_Exc):
    """Exception when used several unsuitable levels in nesting!

    EXAMPLE:
        VictimBase = SingletonWMetaCall
        setattr(VictimBase, "attr", 0)
        class Victim1(VictimBase):
            attr = 1

        assert VictimBase().attr == 0
        try:
            assert Victim1().attr == 1
        except Exc_SingletonDifferentNestingLevels:
            pass
        else:
            assert False

    MAIN RULES:
    1. always instantiate only last Classes in your tree project!


    SPECIALLY CREATED FOR
    ---------------------
    Base_SingletonManager
    """
    pass


# =====================================================================================================================
if __name__ == '__main__':
    # WITH RAISING =====================================
    # REASON --------------
    assert bool(Exception(0)) is True
    assert bool(Exception(False)) is True

    # SOLUTION --------------
    assert bool(Base_Exc(0)) is False
    assert bool(Base_Exc(False)) is False

    # NO RAISING =====================================
    Base_Exc(0, 1, 2, 3)
    Warn(0, 1)


# =====================================================================================================================
